#!/usr/bin/env python
# -*- coding: utf-8 -*-

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""This module provides a set of basic Taurus widgets based on QLed"""
from __future__ import absolute_import

from builtins import str
from builtins import object

import weakref
import collections

from taurus.external.qt import Qt

from taurus.core import DataFormat, AttrQuality, DataType

from taurus.qt.qtgui.base import TaurusBaseWidget
from .qled import QLed


__all__ = ["TaurusLed"]

__docformat__ = 'restructuredtext'

_QT_PLUGIN_INFO = {
    'module': 'taurus.qt.qtgui.display',
    'group': 'Taurus Display',
    'icon': "designer:ledgreen.png",
}


class _TaurusLedController(object):

    #            key     status,     color, inTrouble
    LedMap = {True: (True,    "green",    False),
              False: (False,    "black",    False),
              None: (False,    "black",     True)}

    LedQualityMap = {
        AttrQuality.ATTR_ALARM: (True,   "orange",    False),
        AttrQuality.ATTR_CHANGING: (True,     "blue",    False),
        AttrQuality.ATTR_INVALID: (True,      "red",    False),
        AttrQuality.ATTR_VALID: (True,    "green",    False),
        AttrQuality.ATTR_WARNING: (True,   "orange",    False),
        None: (False,    "black",     True)}

    def __init__(self, widget):
        self._widget = weakref.ref(widget)

    def widget(self):
        return self._widget()

    def modelObj(self):
        return self.widget().getModelObj()

    def value(self):
        widget, obj = self.widget(), self.modelObj()
        fgRole = widget.fgRole
        value = None
        if fgRole == 'rvalue':
            value = obj.rvalue
        elif fgRole == 'wvalue':
            value = obj.wvalue
        elif fgRole == 'quality':
            return obj.quality

        # handle 1D and 2D values
        if obj.data_format is not DataFormat._0D:
            idx = widget.getModelIndexValue()
            if idx:
                for i in idx:
                    value = value[i]

        return bool(value)

    def usePreferedColor(self, widget):
        return True

    def handleEvent(self, evt_src, evt_type, evt_value):
        self.update()

    def update(self):
        widget = self.widget()

        self._updateDisplay(widget)
        self._updateToolTip(widget)

    def _updateDisplay(self, widget):
        key = None
        try:
            key = self.value()
        except Exception:
            pass
        ledMap = self.LedMap
        if widget.fgRole == 'quality':
            ledMap = self.LedQualityMap
        try:
            status, color, trouble = ledMap[key]
        except:
            status, color, trouble = False, "red", True
        if self.usePreferedColor(widget):
            if status:
                color = widget.onColor
            else:
                color = widget.offColor
        widget.ledStatus = status
        widget.ledColor = color
        if trouble:
            widget.setAutoFillBackground(True)
            bg_brush = Qt.QBrush(Qt.Qt.BDiagPattern)
            palette = widget.palette()
            palette.setBrush(Qt.QPalette.Window, bg_brush)
            palette.setBrush(Qt.QPalette.Base, bg_brush)
            widget.setPalette(palette)
        else:
            widget.setAutoFillBackground(False)

    def _updateToolTip(self, widget):
        widget.setToolTip(widget.getFormatedToolTip())


class _TaurusLedControllerBool(_TaurusLedController):

    def usePreferedColor(self, widget):
        # use prefered widget color if representing the boolean read or write
        # value. If representing the quality, use the quality map
        return widget.fgRole != 'quality'

try:
    from taurus.core.tango import DevState  # TODO: Tango-centric
    class _TaurusLedControllerState(_TaurusLedController):

        #                key      status,       color, inTrouble
        LedMap = {DevState.ON: (True,    "green",    False),
                  DevState.OFF: (False,    "black",    False),
                  DevState.CLOSE: (True,    "white",    False),
                  DevState.OPEN: (True,    "green",    False),
                  DevState.INSERT: (True,    "green",    False),
                  DevState.EXTRACT: (True,    "green",    False),
                  DevState.MOVING: (True,     "blue",    False),
                  DevState.STANDBY: (True,   "yellow",    False),
                  DevState.FAULT: (True,      "red",    False),
                  DevState.INIT: (True,   "yellow",    False),
                  DevState.RUNNING: (True,     "blue",    False),
                  DevState.ALARM: (True,   "orange",    False),
                  DevState.DISABLE: (True,  "magenta",    False),
                  DevState.UNKNOWN: (False,    "black",    False),
                  None: (False,    "black",     True)}

        def value(self):
            widget, obj = self.widget(), self.modelObj()
            fgRole = widget.fgRole
            value = None
            if fgRole == 'rvalue':
                value = obj.rvalue
            elif fgRole == 'wvalue':
                value = obj.wvalue
            elif fgRole == 'quality':
                value = obj.quality
            return value

        def usePreferedColor(self, widget):
            # never use prefered widget color. Use always the map
            return False
except:
    pass


class _TaurusLedControllerDesignMode(_TaurusLedController):

    def _updateDisplay(self, widget):
        widget.ledStatus = True
        if widget.ledStatus:
            widget.ledColor = widget.onColor
        else:
            widget.ledColor = widget.offColor
        widget.setAutoFillBackground(False)

    def _updateToolTip(self, widget):
        widget.setToolTip("Design mode TaurusLed")


class TaurusLed(QLed, TaurusBaseWidget):
    """A widget designed to represent with a LED image the state of a device,
    the value of a boolean attribute or the quality of an attribute."""

    DefaultModelIndex = None
    DefaultFgRole = 'rvalue'
    DefaultOnColor = "green"
    DefaultOffColor = "black"

    _deprecatedRoles = dict(value='rvalue', w_value='wvalue')

    def __init__(self, parent=None, designMode=False):
        name = self.__class__.__name__
        self._designMode = designMode
        self._modelIndex = self.DefaultModelIndex
        self._modelIndexStr = ''
        self._fgRole = self.DefaultFgRole
        self._onColor = self.DefaultOnColor
        self._offColor = self.DefaultOffColor
        self._controller = None
        self.call__init__wo_kw(QLed, parent)
        self.call__init__(TaurusBaseWidget, name, designMode=designMode)

        # if we are in design mode there will be no events so we force the
        # creation of a controller object
        if self._designMode:
            self.controller().update()

    def _calculate_controller_class(self):
        model = self.getModelObj()

        klass = _TaurusLedController
        if self._designMode:
            klass = _TaurusLedControllerDesignMode
        elif model is None:
            klass = _TaurusLedController
        elif model.isBoolean():
            klass = _TaurusLedControllerBool
        elif model.type == DataType.DevState:
            klass = _TaurusLedControllerState  # TODO: tango-centric
        return klass

    def controller(self):
        ctrl = self._controller
        # if there is a controller object and it is not the base controller...
        if ctrl is not None and not ctrl.__class__ == _TaurusLedController:
            return ctrl

        # if there is a controller object and it is still the same class...
        ctrl_klass = self._calculate_controller_class()
        if ctrl is not None and ctrl.__class__ == ctrl_klass:
            return ctrl

        self._controller = ctrl = ctrl_klass(self)
        return ctrl

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # TaurusBaseWidget overwriting
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def handleEvent(self, evt_src, evt_type, evt_value):
        self.controller().handleEvent(evt_src, evt_type, evt_value)

    def isReadOnly(self):
        return True

    def setModel(self, m):
        # force to build another controller
        self._controller = None
        TaurusBaseWidget.setModel(self, m)

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # QT property definition
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def getFgRole(self):
        return self._fgRole

    def setFgRole(self, fgRole):
        role = self._deprecatedRoles.get(fgRole, fgRole)
        if fgRole != role:
            self.deprecated(rel='4.0', dep='setFgRole(%s)' % fgRole,
                            alt='setFgRole(%s)' % role)
        self._fgRole = str(role)
        self.controller().update()

    def resetFgRole(self):
        self.setFgRole(self.DefaultFgRole)

    def getOnColor(self):
        """Returns the preferred led on color
        :return: led on color
        :rtype: str"""
        return self._onColor

    def setOnColor(self, color):
        """Sets the preferred led on color
        :param status: the new on color
        :type  status: str"""
        color = str(color).lower()
        if not self.isLedColorValid(color):
            raise Exception("Invalid color '%s'" % color)
        self._onColor = color
        self.controller().update()

    def resetOnColor(self):
        """Resets the preferred led on color"""
        self.setOnColor(self.DefaultOnColor)

    def getOffColor(self):
        """Returns the preferred led off color
        :return: led off color
        :rtype: str"""
        return self._offColor

    def setOffColor(self, color):
        """Sets the preferred led off color
        :param status: the new off color
        :type  status: str"""
        color = str(color).lower()
        if not self.isLedColorValid(color):
            raise Exception("Invalid color '%s'" % color)
        self._offColor = color
        self.controller().update()

    def resetOffColor(self):
        """Resets the preferred led color"""
        self.setOffColor(self.DefaultOffColor)

    def getModelIndexValue(self):
        return self._modelIndex

    def getModelIndex(self):
        return self._modelIndexStr

    def setModelIndex(self, modelIndex):
        mi = str(modelIndex)
        if len(mi) == 0:
            self._modelIndex = None
        else:
            try:
                mi_value = eval(str(mi))
            except:
                return
            if type(mi_value) == int:
                mi_value = mi_value,
            if not isinstance(mi_value, collections.Sequence):
                return
            self._modelIndex = mi_value
        self._modelIndexStr = mi
        self.controller().update()

    def resetModelIndex(self):
        self.setModelIndex(self.DefaultModelIndex)

    @classmethod
    def getQtDesignerPluginInfo(cls):
        d = TaurusBaseWidget.getQtDesignerPluginInfo()
        d.update(_QT_PLUGIN_INFO)
        return d

    #: This property holds the unique URI string representing the model name
    #: with which this widget will get its data from. The convention used for
    #: the string can be found :ref:`here <model-concept>`.
    #:
    #: **Access functions:**
    #:
    #:     * :meth:`TaurusBaseWidget.getModel`
    #:     * :meth:`TaurusLabel.setModel`
    #:     * :meth:`TaurusBaseWidget.resetModel`
    #:
    #: .. seealso:: :ref:`model-concept`
    model = Qt.pyqtProperty("QString", TaurusBaseWidget.getModel, setModel,
                            TaurusBaseWidget.resetModel)

    #: This property holds whether or not this widget should search in the
    #: widget hierarchy for a model prefix in a parent widget.
    #:
    #: **Access functions:**
    #:
    #:     * :meth:`TaurusBaseWidget.getUseParentModel`
    #:     * :meth:`TaurusBaseWidget.setUseParentModel`
    #:     * :meth:`TaurusBaseWidget.resetUseParentModel`
    #:
    #: .. seealso:: :ref:`model-concept`
    useParentModel = Qt.pyqtProperty("bool", TaurusBaseWidget.getUseParentModel,
                                     TaurusBaseWidget.setUseParentModel,
                                     TaurusBaseWidget.resetUseParentModel)

    #: This property holds the index inside the model value that should be
    #: displayed
    #:
    #: **Access functions:**
    #:
    #:     * :meth:`TaurusLed.getModelIndex`
    #:     * :meth:`TaurusLed.setModelIndex`
    #:     * :meth:`TaurusLed.resetModelIndex`
    #:
    #: .. seealso:: :ref:`model-concept`
    modelIndex = Qt.pyqtProperty("QString", getModelIndex, setModelIndex,
                                 resetModelIndex)

    #: This property holds the foreground role.
    #: Valid values are:
    #:
    #:     #. 'value' - the value is used
    #:     #. 'w_value' - the write value is used
    #:     #. 'quality' - the quality is used
    #:
    #: **Access functions:**
    #:
    #:     * :meth:`TaurusLed.getFgRole`
    #:     * :meth:`TaurusLed.setFgRole`
    #:     * :meth:`TaurusLed.resetFgRole`
    fgRole = Qt.pyqtProperty("QString", getFgRole, setFgRole,
                             resetFgRole, doc="foreground role")

    #: This property holds the preferred led color
    #: This value is used for the cases where the model value does not contain
    #: enough information to distinguish between different On colors.
    #: For example, a bool attribute, when it is False it is displayed with the
    #: off led but when it is true it may be displayed On in any color. The
    #: prefered color would be used in this case.
    #:
    #: **Access functions:**
    #:
    #:     * :meth:`TaurusLed.getOnColor`
    #:     * :meth:`TaurusLed.setOnColor`
    #:     * :meth:`TaurusLed.resetOnColor`
    onColor = Qt.pyqtProperty("QString", getOnColor, setOnColor, resetOnColor,
                              doc="preferred led On color")

    #: This property holds the preferred led color
    #: This value is used for the cases where the model value does not contain
    #: enough information to distinguish between different Off colors.
    #: For example, a bool attribute, when it is False it is displayed with the
    #: off led but when it is true it may be displayed On in any color. The
    #: prefered color would be used in this case.
    #:
    #: **Access functions:**
    #:
    #:     * :meth:`TaurusLed.getOffColor`
    #:     * :meth:`TaurusLed.setOffColor`
    #:     * :meth:`TaurusLed.resetOffColor`
    offColor = Qt.pyqtProperty("QString", getOffColor, setOffColor, resetOffColor,
                               doc="preferred led Off color")


def demo():
    "Led"
    from .demo import taurusleddemo
    return taurusleddemo.main()


def main():
    import sys
    import taurus.qt.qtgui.application
    Application = taurus.qt.qtgui.application.TaurusApplication

    app = Application.instance()
    owns_app = app is None

    if owns_app:
        import taurus.core.util.argparse
        parser = taurus.core.util.argparse.get_taurus_parser()
        parser.usage = "%prog [options] <full_attribute_name(s)>"
        app = Application(sys.argv, cmd_line_parser=parser,
                          app_name="Taurus led demo", app_version="1.0",
                          org_domain="Taurus", org_name="Tango community")

    args = app.get_command_line_args()

    if len(args) == 0:
        w = demo()
    else:
        models = map(str.lower, args)

        w = Qt.QWidget()
        layout = Qt.QGridLayout()
        w.setLayout(layout)
        for model in models:
            led = TaurusLed()
            led.model = model
            layout.addWidget(led)
    w.show()

    if owns_app:
        sys.exit(app.exec_())
    else:
        return w

if __name__ == '__main__':
    main()



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""
This module provides a set of basic taurus widgets based on QAbstractSpinBox
"""

from __future__ import absolute_import

from taurus.external.qt import Qt

from .tauruslineedit import TaurusValueLineEdit
from taurus.qt.qtgui.icon import getStandardIcon
from taurus.core.units import Quantity
from taurus.qt.qtgui.util import PintValidator


__all__ = ["TaurusValueSpinBox", "TaurusValueSpinBoxEx"]

__docformat__ = 'restructuredtext'


class TaurusValueSpinBox(Qt.QAbstractSpinBox):

    modelChanged = Qt.pyqtSignal('const QString &')

    def __init__(self, qt_parent=None, designMode=False):
        Qt.QAbstractSpinBox.__init__(self, qt_parent)

        # Overwrite not to show quality by default
        self._showQuality = False

        lineEdit = TaurusValueLineEdit(designMode=designMode)
        lineEdit.setValidator(PintValidator(self))
        lineEdit.setEnableWheelEvent(True)
        self.setLineEdit(lineEdit)
        self.setAccelerated(True)

    def __getattr__(self, name):
        return getattr(self.lineEdit(), name)

    def setValue(self, v):
        self.lineEdit().setValue(v)

    def getValue(self):
        return self.lineEdit().getValue()

    def keyPressEvent(self, evt):
        return self.lineEdit().keyPressEvent(evt)

    def wheelEvent(self, evt):
        return self.lineEdit().wheelEvent(evt)

    # ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~
    # Mandatory overload from QAbstractSpinBox
    # ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~

    def stepBy(self, steps):
        return self.lineEdit()._stepBy(steps)

    def stepEnabled(self):
        ret = Qt.QAbstractSpinBox.StepEnabled(Qt.QAbstractSpinBox.StepNone)
        if self.getModelObj() is None:
            return ret
        if self.getModelObj().getValueObj() is None:
            return ret

        curr_val = self.getValue()

        if curr_val is None:
            return ret

        ss = self._getSingleStepQuantity()

        if self.validate(str(curr_val + ss), 0)[0] == Qt.QValidator.Acceptable:
            ret |= Qt.QAbstractSpinBox.StepUpEnabled
        if self.validate(str(curr_val - ss), 0)[0] == Qt.QValidator.Acceptable:
            ret |= Qt.QAbstractSpinBox.StepDownEnabled
        return ret

    # ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~
    # Overload from QAbstractSpinBox
    # ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~

    def validate(self, input, pos):
        """Overloaded to use the current validator from the TaurusValueLineEdit,
        instead of the default QAbstractSpinBox validator. If no validator is
        set in the LineEdit, it falls back to the original behaviour
        """
        val = self.lineEdit().validator()
        if val is None:
            return Qt.QAbstractSpinBox.validate(self, input, pos)
        return val.validate(input, pos)

    # ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~
    # Model related methods
    # ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~

    def setModel(self, model):
        self.lineEdit().setModel(model)

    def getModel(self):
        return self.lineEdit().getModel()

    def resetModel(self):
        return self.lineEdit().resetModel()

    def setUseParentModel(self, model):
        self.lineEdit().setUseParentModel(model)

    def getUseParentModel(self):
        return self.lineEdit().getUseParentModel()

    def resetUseParentModel(self):
        return self.lineEdit().resetUseParentModel()

    def setAutoApply(self, model):
        self.lineEdit().setAutoApply(model)

    def getAutoApply(self):
        return self.lineEdit().getAutoApply()

    def resetAutoApply(self):
        return self.lineEdit().resetAutoApply()

    def setForcedApply(self, model):
        self.lineEdit().setForcedApply(model)

    def getForcedApply(self):
        return self.lineEdit().getForcedApply()

    def resetForcedApply(self):
        return self.lineEdit().resetForcedApply()

    def getSingleStep(self):
        return self.lineEdit().getSingleStep()

    @Qt.pyqtSlot(float)
    def setSingleStep(self, step):
        self.lineEdit().setSingleStep(step)

    def resetSingleStep(self):
        self.lineEdit().resetSingleStep()

    def _getSingleStepQuantity(self):
        """
        :return: (Quantity) returns a single step with the units of the current
                 value
        """
        value = self.getValue()
        if value is None:
            return None
        return Quantity(self.getSingleStep(), value.units)

    @classmethod
    def getQtDesignerPluginInfo(cls):
        return {
            'group': 'Taurus Input',
            'icon': 'designer:spinbox.png',
            'module': 'taurus.qt.qtgui.input',
            'container': False}

    singleStep = Qt.pyqtProperty("double", getSingleStep, setSingleStep,
                                 resetSingleStep)

    model = Qt.pyqtProperty("QString", getModel, setModel, resetModel)

    useParentModel = Qt.pyqtProperty("bool", getUseParentModel,
                                     setUseParentModel, resetUseParentModel)

    autoApply = Qt.pyqtProperty("bool", getAutoApply, setAutoApply,
                                resetAutoApply)

    forcedApply = Qt.pyqtProperty("bool", getForcedApply, setForcedApply,
                                  resetForcedApply)


_S = """
QSpinBox::up-button {
    border-width: 0px;
}
QPushButton {
    min-width: 6px;
    min-height: 10px;
    /*border-style: solid;
    border-width: 1px;
    border-color: black;
    border-top-right-radius: 2px;
    border-bottom-right-radius: 2px;*/
    margin: 0px;
    padding: 0px;
}
"""


class TaurusValueSpinBoxEx(Qt.QWidget):

    def __init__(self, qt_parent=None, designMode=False):
        Qt.QWidget.__init__(self, qt_parent)
        layout = Qt.QGridLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)
        self.setLayout(layout)
        self.setStyleSheet(_S)
        self.__dict__['spinBox'] = spin = TaurusValueSpinBox(
            qt_parent=self, designMode=designMode)
        self.__dict__['sliderButton1'] = b1 = Qt.QToolButton(self)
        self.__dict__['sliderButton2'] = b2 = Qt.QToolButton(self)
        b1.setIcon(getStandardIcon(Qt.QStyle.SP_TitleBarShadeButton, b1))
        b2.setIcon(getStandardIcon(Qt.QStyle.SP_TitleBarUnshadeButton, b2))
        layout.addWidget(spin, 0, 0, 2, 1)
        layout.addWidget(b1, 0, 1, 1, 1, Qt.Qt.AlignBottom)
        layout.addWidget(b2, 1, 1, 1, 1, Qt.Qt.AlignTop)
        layout.setColumnStretch(0, 1)
        layout.setColumnStretch(1, 0)

        policy = self.sizePolicy()
        policy.setHorizontalPolicy(Qt.QSizePolicy.Minimum)
        policy.setVerticalPolicy(Qt.QSizePolicy.Fixed)
        self.setSizePolicy(policy)

    def __getattr__(self, name):
        return getattr(self.spinBox, name)

    def __setattr__(self, name, value):
        setattr(self.spinBox, name, value)


def main():
    import sys
    import taurus.qt.qtgui.application
    Application = taurus.qt.qtgui.application.TaurusApplication

    app = Application.instance()
    owns_app = app is None

    if owns_app:
        import taurus.core.util.argparse
        parser = taurus.core.util.argparse.get_taurus_parser()
        parser.usage = "%prog [options] <full_attribute_name(s)>"
        app = Application(sys.argv, cmd_line_parser=parser,
                          app_name="Taurus spinbox demo", app_version="1.0",
                          org_domain="Taurus", org_name="Tango community")

    args = app.get_command_line_args()

    if len(args) == 0:
        w = TaurusValueSpinBox()
        w.setModel('sys/tg_test/1/double_scalar')
        w.resize(300, 50)
    else:
        w = Qt.QWidget()
        layout = Qt.QGridLayout()
        w.setLayout(layout)
        for model in args:
            label = TaurusValueSpinBox()
            label.setModel(model)
            layout.addWidget(label)
        w.resize(300, 50)
    w.show()

    if owns_app:
        sys.exit(app.exec_())
    else:
        return w


if __name__ == "__main__":
    main()
#     import sys
#     from taurus.qt.qtgui.application import TaurusApplication
#     app = TaurusApplication()

#     w = TaurusValueSpinBox()
#     w.setModel('sys/tg_test/1/double_scalar')
#     w.resize(300, 50)
#     w.show()

#     sys.exit(app.exec_())



#!/usr/bin/env python
#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################


"""Unit tests for Taurus Label"""

import unittest
from taurus.external.qt import Qt
from taurus.test import insertTest
from taurus.qt.qtgui.test import GenericWidgetTestCase, BaseWidgetTestCase
from taurus.qt.qtgui.display import TaurusLabel
from taurus.qt.qtgui.container import TaurusWidget
from taurus.core.tango.test import TangoSchemeTestLauncher
import functools
from taurus.core.util.colors import ATTRIBUTE_QUALITY_DATA, DEVICE_STATE_DATA

DEV_NAME = TangoSchemeTestLauncher.DEV_NAME


class TaurusLabelTest(GenericWidgetTestCase, unittest.TestCase):

    '''
    Generic tests for TaurusLabel.

    .. seealso: :class:`taurus.qt.qtgui.test.base.GenericWidgetTestCase`
    '''
    _klass = TaurusLabel
    modelnames = ['sys/tg_test/1/wave', '', 'eval:1', None]


class Bug169_Test(BaseWidgetTestCase, unittest.TestCase):

    '''
    Test bug169:  See: http://sf.net/p/tauruslib/tickets/65/
                  (aka, Sradana's bug169)


    .. seealso: :class:`taurus.qt.qtgui.test.base.BaseWidgetTestCase`
    '''
    _klass = TaurusLabel

    def setUp(self):
        BaseWidgetTestCase.setUp(self)
        self._widget.setModel('sys/tg_test/1/double_scalar#label')
        self._expectedModelClass = self._widget.getModelClass()
        self._parent = TaurusWidget()
        self._parent.setModel('sys/tg_test/1')
        self._widget.setUseParentModel(True)
        self._widget.setParent(self._parent)

    def test_bug169(self):
        '''Check if setModel works when using parent model'''
        self._widget.setModel('/double_scalar#label')
        self.assertMaxDeprecations(1)

    def test_relativemodelclass(self):
        '''Check consistency in modelClass when using parent model (re: bug169)
        '''
        try:
            self._widget.setModel('/double_scalar#label')
        finally:
            mc = self._widget.getModelClass()
            msg = ('getModelClass() inconsistency:\n expected: %s\n got: %s' %
                   (self._expectedModelClass, mc))
            self.assertEqual(self._expectedModelClass, mc, msg)
        self.assertMaxDeprecations(1)


# ------------------------------------------------------------------------------
# Check bck-compat with pre-tep14  FgRoles: value, w_value, state,
# quality, none
testOldFgroles = functools.partial(insertTest, helper_name='text', maxdepr=1,
                                   model='tango:' + DEV_NAME + '/double_scalar')


@testOldFgroles(fgRole='value', expected='1.23 mm')
@testOldFgroles(fgRole='w_value', expected='0.00 mm')
@testOldFgroles(fgRole='state', expected='Ready')
@testOldFgroles(fgRole='quality', expected='ATTR_VALID')
@testOldFgroles(fgRole='none', expected='')
# ------------------------------------------------------------------------------
@insertTest(helper_name='text',
            model='tango:' + DEV_NAME + '/double_image#rvalue[1,::2]',
            expected='[ 1.23  1.23] mm')
@insertTest(helper_name='text',
            model='tango:' + DEV_NAME + '/double_spectrum#rvalue[1]',
            expected='1.23 mm')
@insertTest(helper_name='text',
            model='tango:' + DEV_NAME + '/double_image',
            modelIndex=(1,1),
            expected='1.23 mm')
@insertTest(helper_name='text',
            model='tango:' + DEV_NAME + '/double_spectrum',
            modelIndex=1,
            expected='1.23 mm')
@insertTest(helper_name='text',
            model='tango:' + DEV_NAME + '/double_scalar#state',
            expected='Ready')
@insertTest(helper_name='text',
            model='tango:' + DEV_NAME + '/double_scalar#rvalue',
            fgRole='label',
            expected='double_scalar')
@insertTest(helper_name='text',
            model='tango:' + DEV_NAME + '/double_scalar',
            fgRole='label',
            expected='double_scalar')
@insertTest(helper_name='text',
            model='tango:' + DEV_NAME + '/double_scalar#label',
            expected='double_scalar')
# ------------------------------------------------------------------------------
# Check bck-compat with pre-tep14  BgRoles: state, quality, none
@insertTest(helper_name='bgRole',
            model='tango:' + DEV_NAME + '/float_scalar_ro',
            bgRole='none',
            expected=Qt.QColor(Qt.Qt.transparent).getRgb()[:3])
@insertTest(helper_name='bgRole',
            model='tango:' + DEV_NAME + '/float_scalar_ro',
            bgRole='state',
            expected=DEVICE_STATE_DATA["TaurusDevState.Ready"][1:4])
@insertTest(helper_name='bgRole',
            model='tango:' + DEV_NAME + '/float_scalar_ro',
            bgRole='quality',
            expected=ATTRIBUTE_QUALITY_DATA["ATTR_VALID"][1:4])
@insertTest(helper_name='bgRole',
            model='tango:' + DEV_NAME + '/float_scalar_ro',
            expected=ATTRIBUTE_QUALITY_DATA["ATTR_VALID"][1:4])
class TaurusLabelTest2(TangoSchemeTestLauncher, BaseWidgetTestCase,
                       unittest.TestCase):
    '''
    Specific tests for TaurusLabel
    '''
    _klass = TaurusLabel

    def text(self, model=None, expected=None, fgRole=None, maxdepr=0,
             modelIndex=None):
        """Check that the label text"""
        self._widget.setModel(model)
        if fgRole is not None:
            self._widget.setFgRole(fgRole)
        if modelIndex is not None:
            self._widget.setModelIndex(modelIndex)
        self.processEvents(repetitions=10, sleep=.1)
        got = str(self._widget.text())
        msg = ('wrong text for "%s":\n expected: %s\n got: %s' %
               (model, expected, got))
        self.assertEqual(got, expected, msg)
        self.assertMaxDeprecations(maxdepr)

    # TODO: fix the bgRole test (they fail sometimes when run by testsuite)
    # ~~~~~~~~~~~~~~~~~~~~~~~
    # FAIL: test_bgRole (...)
    #  expected: (0, 255, 0)
    #  got: (239, 240, 241)  # <-- these values change (unitialized garbage?)
    # ~~~~~~~~~~~~~~~~~~~~~~~
    #@unittest.skip('bgRole tests fail randomly. Skip until fixed.')
    def bgRole(self, model=None, expected=None, bgRole=None, maxdepr=0):
        '''Check that the label text'''
        self._widget.setModel(model)
        self.processEvents(repetitions=50, sleep=.1)
        if bgRole is not None:
            self._widget.setBgRole(bgRole)
        p = self._widget.palette()
        got = p.color(p.Background).getRgb()[:3]  # RGB value of the background
        msg = ('wrong background RGB for "%s":\n expected: %s\n got: %s' %
               (model, expected, got))
        self.assertEqual(got, expected, msg)
        self.assertMaxDeprecations(maxdepr)


def baseFormatter1(dtype, **kwargs):
    return "{:~.1f}"


def baseFormatter2(dtype, **kwargs):
    return dtype.__name__

# -------------------------------------------------------------------------
# Class formatter tests
@insertTest(helper_name='checkClassFormat',
            model='eval:1.2345',
            formatter='>>{}<<',
            expected=">>1.2345<<")
@insertTest(helper_name='checkClassFormat',
            model='eval:Q(5)#rvalue.magnitude',
            formatter=baseFormatter2,
            expected="int")
@insertTest(helper_name='checkClassFormat',
            model='eval:Q("5m")#rvalue.units',
            formatter=baseFormatter2,
            expected="Unit")
@insertTest(helper_name='checkClassFormat',
            model='eval:1.2345',
            formatter=baseFormatter1,
            expected="1.2")
@insertTest(helper_name='checkClassFormat',
            model='eval:"hello"',
            formatter=baseFormatter1,
            expected="hello")
@insertTest(helper_name='checkClassFormat',
            model='eval:"hello"',
            formatter=baseFormatter2,
            expected="str")
@insertTest(helper_name='checkClassFormat',
            model='eval:"hello"',
            formatter=None,
            expected="hello")
@insertTest(helper_name='checkClassFormat',
            model='eval:1.2345',
            formatter='{:~.3f}',
            expected="1.234")
@insertTest(helper_name='checkClassFormat',
            model='eval:1.2345',
            formatter='{:.3f}',
            expected="1.234 dimensionless")
# -------------------------------------------------------------------------
# Instance formatter tests
@insertTest(helper_name='checkInstanceFormat',
            model='eval:1.2345',
            formatter='>>{}<<',
            expected=">>1.2345<<")
@insertTest(helper_name='checkInstanceFormat',
            model='eval:Q(5)#rvalue.magnitude',
            formatter=baseFormatter2,
            expected="int")
@insertTest(helper_name='checkInstanceFormat',
            model='eval:Q("5m")#rvalue.units',
            formatter=baseFormatter2,
            expected="Unit")
@insertTest(helper_name='checkInstanceFormat',
            model='eval:1.2345',
            formatter=baseFormatter1,
            expected="1.2")
@insertTest(helper_name='checkInstanceFormat',
            model='eval:"hello"',
            formatter=baseFormatter1,
            expected="hello")
@insertTest(helper_name='checkInstanceFormat',
            model='eval:"hello"',
            formatter=baseFormatter2,
            expected="str")
@insertTest(helper_name='checkInstanceFormat',
            model='eval:"hello"',
            formatter=None,
            expected="hello")
@insertTest(helper_name='checkInstanceFormat',
            model='eval:1.2345',
            formatter='{:~.3f}',
            expected="1.234")
@insertTest(helper_name='checkInstanceFormat',
            model='eval:1.2345',
            formatter='{:.3f}',
            expected="1.234 dimensionless")
class TaurusLabelFormatTest(BaseWidgetTestCase, unittest.TestCase):
    """
    Specific tests for testing the Formatting API with TaurusLabel
    instances
    """

    _klass = TaurusLabel

    def setUp(self):
        BaseWidgetTestCase.setUp(self)
        # store original class format
        self._origFormatter = self._klass.FORMAT

    def tearDown(self):
        # restore original class format
        self._klass.FORMAT = self._origFormatter

    def checkInstanceFormat(self, model, formatter, expected):
        if formatter is not None:
            self._widget.setFormat(formatter)
        self._widget.setModel(model)
        self.processEvents(repetitions=50, sleep=.1)

        got = self._widget.text()
        msg = ('wrong text for "%s":\n expected: %s\n got: %s' %
               (model, expected, got))
        self.assertEqual(got, expected, msg)

    def checkClassFormat(self, model, formatter, expected):
        self._klass.FORMAT = formatter
        # self._widget was already created by BaseWidgetTestCase.setUp(self)
        # but we need to re-create it to use the class formatter
        self._widget = self._klass(*self.initargs, **self.initkwargs)
        self._widget.setModel(model)
        self.processEvents(repetitions=50, sleep=.1)

        got = self._widget.text()
        msg = ('wrong text for "%s":\n expected: %s\n got: %s' %
               (model, expected, got))
        self.assertEqual(got, expected, msg)


#
# if __name__ == "__main__":
#     unittest.main()



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""
taurusvalue.py:
"""

__all__ = ["TaurusValue", "TaurusValuesFrame", "DefaultTaurusValueCheckBox",
           "DefaultUnitsWidget", "TaurusPlotButton", "TaurusArrayEditorButton",
           "TaurusValuesTableButton", "TaurusValuesTableButton_W",
           "DefaultLabelWidget", "DefaultReadWidgetLabel", "TaurusDevButton",
           "TaurusImageButton"]

__docformat__ = 'restructuredtext'

from future.utils import string_types
from future.builtins import str

import weakref
import re
from taurus.external.qt import Qt
import taurus.core
from taurus.core import DataType, DataFormat, TaurusEventType

from taurus.core.taurusbasetypes import TaurusElementType
from taurus.qt.qtcore.mimetypes import TAURUS_ATTR_MIME_TYPE, TAURUS_DEV_MIME_TYPE, TAURUS_MODEL_MIME_TYPE
from taurus.qt.qtcore.configuration import BaseConfigurableClass
from taurus.qt.qtgui.base import TaurusBaseWidget
from taurus.qt.qtgui.container import TaurusFrame
from taurus.qt.qtgui.display import TaurusLabel
from taurus.qt.qtgui.display import TaurusLed
from taurus.qt.qtgui.input import TaurusValueSpinBox, TaurusValueCheckBox
from taurus.qt.qtgui.input import TaurusWheelEdit, TaurusValueLineEdit
from taurus.qt.qtgui.button import TaurusLauncherButton
from taurus.qt.qtgui.util import TaurusWidgetFactory, ConfigurationMenu
from taurus.qt.qtgui.compact import TaurusReadWriteSwitcher


class DefaultTaurusValueCheckBox(TaurusValueCheckBox):

    def __init__(self, *args):
        TaurusValueCheckBox.__init__(self, *args)
        self.setShowText(False)

    @classmethod
    def getQtDesignerPluginInfo(cls):
        return None


class DefaultLabelWidget(TaurusLabel):
    '''
    The base class used by default for showing the label of a TaurusValue.

    .. note:: It only makes sense to use this class as a part of a TaurusValue,
              since it assumes that it can get a reference to a TaurusValue via
              the :meth:`getTaurusValueBuddy` member
    '''

    _dragEnabled = True

    def __init__(self, *args):
        TaurusLabel.__init__(self, *args)
        self.setAlignment(Qt.Qt.AlignRight)
        self.setSizePolicy(Qt.QSizePolicy.Preferred, Qt.QSizePolicy.Maximum)
        self.setBgRole(None)
        self.autoTrim = False
        self.setStyleSheet(
            'DefaultLabelWidget {border-style: solid; border-width: 1px; border-color: transparent; border-radius: 4px;}')

    def setModel(self, model):
        if model is None or model == '':
            return TaurusLabel.setModel(self, None)
        try:
            config = self.taurusValueBuddy().getLabelConfig()
        except Exception:
            config = '{attr.label}'
        elementtype = self.taurusValueBuddy().getModelType()
        fullname = self.taurusValueBuddy().getModelObj().getFullName()
        if elementtype == TaurusElementType.Attribute:
            config = self.taurusValueBuddy().getLabelConfig()
            TaurusLabel.setModel(self, '%s#%s' % (fullname, config))
        elif elementtype == TaurusElementType.Device:
            devName = self.taurusValueBuddy().getModelObj().getSimpleName()
            TaurusLabel.setModel(self, model)
            self.setText(devName)

    _BCK_COMPAT_TAGS = {'<attr_name>': '{attr.name}',
                        '<attr_fullname>': '{attr.fullname}',
                        '<dev_alias>': '{dev.name}',
                        '<dev_name>': '{dev.name}',
                        '<dev_fullname>': '{dev.fullname}',
                        }

    def getDisplayValue(self, cache=True, fragmentName=None):
        """Reimplementation of getDisplayValue"""
        if self.modelObj is None or fragmentName is None:
            return self.getNoneValue()
        # support bck-compat tags
        for old in re.findall('<.+?>', fragmentName):
            new = self._BCK_COMPAT_TAGS.get(old, '{attr.%s}' % old)
            self.deprecated(dep=old, alt=new)
            fragmentName = fragmentName.replace(old, new)

        return TaurusLabel.displayValue(self, fragmentName)

    def sizeHint(self):
        return Qt.QSize(Qt.QLabel.sizeHint(self).width(), 18)

    def contextMenuEvent(self, event):
        """ The label widget will be used for handling the actions of the whole TaurusValue

        see :meth:`QWidget.contextMenuEvent`"""
        menu = Qt.QMenu(self)
        # @todo: This should be done more Taurus-ish
        menu.addMenu(ConfigurationMenu(self.taurusValueBuddy()))
        if hasattr(self.taurusValueBuddy().writeWidget(followCompact=True), 'resetPendingOperations'):
            r_action = menu.addAction("reset write value", self.taurusValueBuddy(
            ).writeWidget(followCompact=True).resetPendingOperations)
            r_action.setEnabled(self.taurusValueBuddy().hasPendingOperations())
        if self.taurusValueBuddy().isModifiableByUser():
            menu.addAction("Change label",
                           self.taurusValueBuddy()._onChangeLabelText)
            menu.addAction("Change Read Widget",
                           self.taurusValueBuddy().onChangeReadWidget)
            menu.addAction("Set Formatter",
                           self.taurusValueBuddy().onSetFormatter)
            cw_action = menu.addAction(
                "Change Write Widget", self.taurusValueBuddy().onChangeWriteWidget)
            # disable the action if the taurusValue is readonly
            cw_action.setEnabled(not self.taurusValueBuddy().isReadOnly())
            cm_action = menu.addAction("Compact")
            cm_action.setCheckable(True)
            cm_action.setChecked(self.taurusValueBuddy().isCompact())
            cm_action.toggled.connect(self.taurusValueBuddy().setCompact)
        menu.exec_(event.globalPos())
        event.accept()

    def getModelMimeData(self):
        """
        reimplemented to use the taurusValueBuddy model instead of its own
        model
        """
        mimeData = TaurusLabel.getModelMimeData(self)
        _modelname = str(self.taurusValueBuddy().getModelName())
        modelname = _modelname.encode(encoding='utf8')
        mimeData.setData(TAURUS_MODEL_MIME_TYPE, modelname)
        if self.taurusValueBuddy().getModelType() == TaurusElementType.Device:
            mimeData.setData(TAURUS_DEV_MIME_TYPE, modelname)
        elif self.taurusValueBuddy().getModelType() == TaurusElementType.Attribute:
            mimeData.setData(TAURUS_ATTR_MIME_TYPE, modelname)
        return mimeData

    @classmethod
    def getQtDesignerPluginInfo(cls):
        return None


class ExpandingLabel(TaurusLabel):
    '''just a expanding TaurusLabel'''

    def __init__(self, *args):
        TaurusLabel.__init__(self, *args)
        self.setSizePolicy(Qt.QSizePolicy.Expanding, Qt.QSizePolicy.Preferred)


class DefaultReadWidgetLabel(ExpandingLabel):
    """A customised label for the read widget"""

    def setModel(self, m):
        TaurusLabel.setModel(self, m)
        fgrole = 'rvalue'
        model_obj = self.getModelObj()
        if model_obj is None:
            return
        if model_obj.getType() in (DataType.Integer, DataType.Float):
            fgrole += '.magnitude'
        self.setFgRole(fgrole)


class CenteredLed(TaurusLed):
    '''just a centered TaurusLed'''
    DefaultAlignment = Qt.Qt.AlignHCenter | Qt.Qt.AlignVCenter


class DefaultUnitsWidget(TaurusLabel):

    FORMAT = "{}"

    def __init__(self, *args):
        TaurusLabel.__init__(self, *args)
        self.setNoneValue('')
        self.setSizePolicy(Qt.QSizePolicy.Preferred, Qt.QSizePolicy.Maximum)
        self.autoTrim = False
        self.setBgRole(None)
        self.setAlignment(Qt.Qt.AlignLeft)

    def setModel(self, model):
        if model is None or model == '':
            return TaurusLabel.setModel(self, None)
        TaurusLabel.setModel(self, model + "#rvalue.units")

    def sizeHint(self):
        # print "UNITSSIZEHINT:",Qt.QLabel.sizeHint(self).width(),
        # self.minimumSizeHint().width(),
        # Qt.QLabel.minimumSizeHint(self).width()
        return Qt.QSize(Qt.QLabel.sizeHint(self).width(), 24)

    @classmethod
    def getQtDesignerPluginInfo(cls):
        return None


class _AbstractTaurusValueButton(TaurusLauncherButton):
    _deleteWidgetOnClose = True
    _text = 'Show'

    def __init__(self, parent=None, designMode=False):
        TaurusLauncherButton.__init__(
            self, parent=parent, designMode=designMode)
        self.setSizePolicy(Qt.QSizePolicy.Expanding, Qt.QSizePolicy.Maximum)

    @classmethod
    def getQtDesignerPluginInfo(cls):
        return None


class TaurusPlotButton(_AbstractTaurusValueButton):
    '''A button that launches a TaurusPlot'''
    _widgetClassName = 'TaurusPlot'
    _icon = 'designer:qwtplot.png'


class TaurusArrayEditorButton(_AbstractTaurusValueButton):
    '''A button that launches a TaurusArrayEditor'''
    _widgetClassName = 'TaurusArrayEditor'
    _icon = 'designer:arrayedit.png'
    _text = 'Edit'


class TaurusImageButton(_AbstractTaurusValueButton):
    '''A button that launches an ImageDialog'''
    _widgetClassName = 'TaurusImageDialog'
    _icon = 'mimetypes:image-x-generic.svg'


class TaurusValuesTableButton(_AbstractTaurusValueButton):
    '''A button that launches a TaurusValuesTable'''
    _widgetClassName = 'TaurusValuesTable'
    _icon = 'designer:table.png'
    _kwargs = {'defaultWriteMode': 'r'}


class TaurusValuesTableButton_W(TaurusValuesTableButton):
    '''A button that launches a TaurusValuesTable'''
    _text = 'Edit'
    _kwargs = {'defaultWriteMode': 'w'}


class TaurusDevButton(_AbstractTaurusValueButton):
    '''A button that launches a TaurusAttrForm'''
    _widgetClassName = 'TaurusDevicePanel'
    _icon = 'places:folder-remote.svg'
    _text = 'Show Device'


class TaurusStatusLabel(TaurusLabel):
    '''just a taurusLabel but showing the state as its background by default'''

    def __init__(self, parent=None, designMode=False):
        TaurusLabel.__init__(self, parent=parent, designMode=designMode)
        self.setBgRole('state')
        self.setSizePolicy(Qt.QSizePolicy.Expanding, Qt.QSizePolicy.Maximum)

    @classmethod
    def getQtDesignerPluginInfo(cls):
        return None


class TaurusValue(Qt.QWidget, TaurusBaseWidget):
    '''
    Internal TaurusValue class

    .. warning::

        :class:`TaurusValue` (and any derived class from it) should never be instantiated directly.
        It is designed to be instantiated by a :class:`TaurusForm` class, since it
        breaks some conventions on the way it manages layouts of its parent model.
    '''
    _compact = False

    def __init__(self, parent=None, designMode=False, customWidgetMap=None):
        name = self.__class__.__name__
        self.call__init__wo_kw(Qt.QWidget, parent)
        self.call__init__(TaurusBaseWidget, name, designMode=designMode)

        self.__error = False
        self.__modelClass = None
        self._designMode = designMode

        # This is a hack to show something usable when in designMode
        if self._designMode:
            layout = Qt.QHBoxLayout(self)
            dummy = ExpandingLabel()
            layout.addWidget(dummy)
            #dummy.setUseParentModel(True)
            dummy.setModel("#attr_fullname")
            dummy.setPrefixText("< TaurusValue: ")
            dummy.setSuffixText(" >")
        else:
            self.setFixedSize(0, 0)

        self._labelWidget = None
        self._readWidget = None
        self._writeWidget = None
        self._unitsWidget = None
        self._customWidget = None
        self._extraWidget = None

        if customWidgetMap is None:
            customWidgetMap = {}
        self.setCustomWidgetMap(customWidgetMap)

        self.labelWidgetClassID = 'Auto'
        self.readWidgetClassID = 'Auto'
        self.writeWidgetClassID = 'Auto'
        self.unitsWidgetClassID = 'Auto'
        self.customWidgetClassID = 'Auto'
        self.extraWidgetClassID = 'Auto'
        self.setPreferredRow(-1)
        self._row = None

        self._allowWrite = True
        self._minimumHeight = None
        self._labelConfig = '{attr.label}'
        self.setModifiableByUser(False)

        if parent is not None:
            self.setParent(parent)

        self.registerConfigProperty(
            self.getLabelConfig, self.setLabelConfig, 'labelConfig')
        self.registerConfigProperty(self.isCompact, self.setCompact, 'compact')

    def setVisible(self, visible):
        for w in (self.labelWidget(), self.readWidget(), self.writeWidget(), self.unitsWidget(), self.customWidget(), self.extraWidget()):
            if w is not None:
                w.setVisible(visible)
        Qt.QWidget.setVisible(self, visible)

    def labelWidget(self):
        '''Returns the label widget'''
        return self._labelWidget

    def readWidget(self, followCompact=False):
        '''
        Returns the read widget. If followCompact=True, and compact mode is
        used, it returns the switcher's readWidget instead of the switcher
        itself.
        '''
        if followCompact and self.isCompact():
            return self._readWidget.readWidget
        return self._readWidget

    def writeWidget(self, followCompact=False):
        '''
        Returns the write widget. If followCompact=True, and compact mode is
        used, it returns the switcher's writeWidget instead of None.
        '''
        if followCompact and self.isCompact():
            return getattr(self._readWidget, 'writeWidget', None)
        return self._writeWidget

    def unitsWidget(self):
        '''Returns the units widget'''
        return self._unitsWidget

    def customWidget(self):
        '''Returns the custom widget'''
        return self._customWidget

    def extraWidget(self):
        '''Returns the extra widget'''
        return self._extraWidget

    def setParent(self, parent):

        # make sure that the parent has a QGriLayout
        pl = parent.layout()
        if pl is None:
            pl = Qt.QGridLayout(parent)  # creates AND sets the parent layout
        if not isinstance(pl, Qt.QGridLayout):
            raise ValueError(
                'layout must be a QGridLayout (got %s)' % type(pl))

        if self._row is None:
            # @TODO we should check that the Preferred row is empty in pl
            self._row = self.getPreferredRow()
            if self._row < 0:
                self._row = pl.rowCount()
        # print 'ROW:',self, self.getRow()

        # insert self into the 0-column
        # this widget is invisible (except in design mode)
        pl.addWidget(self, self._row, 0)

        # Create/update the subwidgets (this also inserts them in the layout)
        if not self._designMode:  # in design mode, no subwidgets are created
            self.updateLabelWidget()
            self.updateReadWidget()
            self.updateWriteWidget()
            self.updateUnitsWidget()
            self.updateExtraWidget()

#        self.updateCustomWidget()

        # do the base class stuff too
        Qt.QWidget.setParent(self, parent)

    def onSetFormatter(self):
        """
        Reimplemented to call onSetFormatter of the read widget (if provided)
        """
        rw = self.readWidget(followCompact=True)
        if hasattr(rw, 'onSetFormatter'):
            return rw.onSetFormatter()

    def setFormat(self, format):
        """
        Reimplemented to call setFormat of the read widget (if provided)
        """
        TaurusBaseWidget.setFormat(self, format)
        try:
            rw = self.readWidget(followCompact=True)
        except AttributeError:
            return
        if hasattr(rw, 'setFormat'):
            rw.setFormat(format)

    def getAllowWrite(self):
        return self._allowWrite

    @Qt.pyqtSlot(bool)
    def setAllowWrite(self, mode):
        self._allowWrite = mode

    def resetAllowWrite(self):
        self._allowWrite = True

    def getPreferredRow(self):
        return self._preferredRow

    @Qt.pyqtSlot(int)
    def setPreferredRow(self, row):
        self._preferredRow = row

    def resetPreferredRow(self):
        self.setPreferredRow(-1)

    def getRow(self):
        return self._row

    def setMinimumHeight(self, minimumHeight):
        self._minimumHeight = minimumHeight

    def minimumHeight(self):
        return self._minimumHeight

    def getDefaultLabelWidgetClass(self):
        return DefaultLabelWidget

    def getDefaultReadWidgetClass(self, returnAll=False):
        '''
        Returns the default class (or classes) to use as read widget for the
        current model.

        :param returnAll: (bool) if True, the return value is a list of valid
                          classes instead of just one class

        :return: (class or list<class>) the default class  to use for the read
                 widget (or, if returnAll==True, a list of classes that can show
                 the attribute ). If a list is returned, it will be loosely
                 ordered by preference, being the first element always the
                 default one.
        '''
        modelobj = self.getModelObj()
        if modelobj is None:
            if returnAll:
                return [DefaultReadWidgetLabel]
            else:
                return DefaultReadWidgetLabel

        modeltype = self.getModelType()
        if modeltype == TaurusElementType.Attribute:
            # The model is an attribute
            # print "---------ATTRIBUTE OBJECT:----------\n",modelobj.read()
            try:
                if modelobj.isBoolean():
                    result = [CenteredLed, DefaultReadWidgetLabel]
            except:
                pass
            if modelobj.data_format == DataFormat._0D:
                if modelobj.type == DataType.Boolean:
                    result = [CenteredLed, DefaultReadWidgetLabel]
                elif modelobj.type == DataType.DevState:
                    result = [CenteredLed, DefaultReadWidgetLabel]
                elif str(self.getModel()).lower().endswith('/status'):  # @todo: tango-centric!!
                    result = [TaurusStatusLabel, DefaultReadWidgetLabel]
                else:
                    result = [DefaultReadWidgetLabel]
            elif modelobj.data_format == DataFormat._1D:
                if modelobj.type in (DataType.Float, DataType.Integer):
                    result = [TaurusPlotButton,
                              TaurusValuesTableButton, DefaultReadWidgetLabel]
                else:
                    result = [TaurusValuesTableButton, DefaultReadWidgetLabel]
            elif modelobj.data_format == DataFormat._2D:
                if modelobj.type in (DataType.Float, DataType.Integer):
                    try:
                        # unused import but useful to determine if
                        # TaurusImageButton should be added
                        from taurus.qt.qtgui.extra_guiqwt import TaurusImageDialog
                        result = [TaurusImageButton,
                                  TaurusValuesTableButton, DefaultReadWidgetLabel]
                    except ImportError:
                        result = [TaurusValuesTableButton,
                                  DefaultReadWidgetLabel]
                else:
                    result = [TaurusValuesTableButton, DefaultReadWidgetLabel]
            else:
                self.warning('Unsupported attribute type %s' % modelobj.type)
                result = None

        elif modeltype == TaurusElementType.Device:
            result = [TaurusDevButton]
        else:
            msg = "Unsupported model type ('%s')" % modeltype
            self.warning(msg)
            raise ValueError(msg)

        if returnAll:
            return result
        else:
            return result[0]

    def getDefaultWriteWidgetClass(self, returnAll=False):
        '''
        Returns the default class (or classes) to use as write widget for the
        current model.

        :param returnAll: (bool) if True, the return value is a list of valid
                          classes instead of just one class

        :return: (class or list<class>) the default class  to use for the write
                 widget (or, if returnAll==True, a list of classes that can show
                 the attribute ). If a list is returned, it will be loosely
                 ordered by preference, being the first element always the
                 default one.
        '''
        modelclass = self.getModelClass()
        if self.isReadOnly() or (modelclass and modelclass.getTaurusElementType() != TaurusElementType.Attribute):
            if returnAll:
                return []
            else:
                return None
        modelobj = self.getModelObj()
        if modelobj is None:
            if returnAll:
                return [TaurusValueLineEdit]
            else:
                return TaurusValueLineEdit
        modelType = modelobj.getType()
        if modelobj.data_format == DataFormat._0D:
            if modelType == DataType.Boolean:
                result = [DefaultTaurusValueCheckBox, TaurusValueLineEdit]
            else:
                result = [TaurusValueLineEdit,
                          TaurusValueSpinBox, TaurusWheelEdit]
        elif modelobj.data_format == DataFormat._1D:
            if modelType in (DataType.Float, DataType.Integer):
                result = [TaurusArrayEditorButton,
                          TaurusValuesTableButton_W, TaurusValueLineEdit]
            else:
                result = [TaurusValuesTableButton_W, TaurusValueLineEdit]
        elif modelobj.data_format == DataFormat._2D:
            result = [TaurusValuesTableButton_W]
        else:
            self.debug('Unsupported attribute type for writing: %s' %
                       str(DataType.whatis(modelType)))
            result = [None]

        if returnAll:
            return result
        else:
            return result[0]

    def getDefaultUnitsWidgetClass(self):
        #        if self.getModelClass() != taurus.core.taurusattribute.TaurusAttribute:
        #            return DefaultUnitsWidget
        return DefaultUnitsWidget

    def getDefaultCustomWidgetClass(self):
        modelclass = self.getModelClass()
        if modelclass and modelclass.getTaurusElementType() != TaurusElementType.Device:
            return None
        try:
            key = self.getModelObj().getDeviceProxy().info().dev_class  # TODO: Tango-centric
        except:
            return None
        return self.getCustomWidgetMap().get(key, None)

    def getDefaultExtraWidgetClass(self):
        return None

    def setCustomWidgetMap(self, cwmap):
        '''Sets a map map for custom widgets.

        :param cwmap: (dict<str,Qt.QWidget>) a dictionary whose keys are device
                      class strings (see :class:`PyTango.DeviceInfo`) and
                      whose values are widget classes to be used
        '''
        self._customWidgetMap = cwmap

    def getCustomWidgetMap(self):
        '''Returns the map used to create custom widgets.

        :return: (dict<str,Qt.QWidget>) a dictionary whose keys are device
                 type strings (i.e. see :class:`PyTango.DeviceInfo`) and whose
                 values are widgets to be used
        '''
        return self._customWidgetMap

    def onChangeLabelConfig(self):
        self.deprecated(msg="onChangeLabelConfig is deprecated", rel="Jan2018")
        self._onChangeLabelText()

    def _onChangeLabelText(self):
        keys = ['{attr.label}', '{attr.name}', '{attr.fullname}', '{dev.name}',
                '{dev.fullname}']

        try:
            current = keys.index(self.labelConfig)
        except:
            current = len(keys)
            keys.append(self.labelConfig)

        msg = 'Choose the label format. \n' + \
              'You may use Python format() syntax. The TaurusDevice object\n' + \
              'can be referenced as "dev" and the TaurusAttribute object\n' + \
              'as "attr"'

        labelConfig, ok = Qt.QInputDialog.getItem(self, 'Change Label', msg,
                                                  keys, current, True)
        if ok:
            self.labelConfig = labelConfig

    def onChangeReadWidget(self):
        classnames = ['None', 'Auto'] + \
            [c.__name__ for c in self.getDefaultReadWidgetClass(
                returnAll=True)]
        cname, ok = Qt.QInputDialog.getItem(
            self, 'Change Read Widget', 'Choose a new read widget class', classnames, 1, True)
        if ok:
            self.setReadWidgetClass(str(cname))

    def onChangeWriteWidget(self):
        classnames = ['None', 'Auto'] + \
            [c.__name__ for c in self.getDefaultWriteWidgetClass(
                returnAll=True)]
        cname, ok = Qt.QInputDialog.getItem(
            self, 'Change Write Widget', 'Choose a new write widget class', classnames, 1, True)
        if ok:
            self.setWriteWidgetClass(str(cname))

    def _destroyWidget(self, widget):
        '''get rid of a widget in a safe way'''
        widget.hide()
        widget.setParent(None)
        if hasattr(widget, 'setModel'):
            widget.setModel(None)
        # COULD NOT INVESTIGATE DEEPER, BUT THE STARTUP-HANGING
        # HAPPENS WITH SOME SIGNALS RELATED WITH THE LINEEDIT...
        # MAYBE OTHER 'WRITE WIDGETS' HAVE THE SAME PROBLEM ?!?!?!
        if isinstance(widget, Qt.QLineEdit):
            widget.blockSignals(True)
        # THIS HACK REDUCES THE STARTUP-HANGING RATE
        widget.deleteLater()

    def _newSubwidget(self, oldWidget, newClass):
        '''eliminates oldWidget and returns a new one.
        If newClass is None, None is returned
        If newClass is the same as the olWidget class, nothing happens'''
        if oldWidget.__class__ == newClass:
            return oldWidget
        if oldWidget is not None:
            self._destroyWidget(oldWidget)
        if newClass is None:
            result = None
        else:
            result = newClass()
        return result

    def labelWidgetClassFactory(self, classID):
        if self._customWidget is not None:
            return None
        if classID is None or classID == 'None':
            return None
        classID = globals().get(classID, classID)
        if isinstance(classID, type):
            return classID
        elif str(classID) == 'Auto':
            return self.getDefaultLabelWidgetClass()
        else:
            return TaurusWidgetFactory().getWidgetClass(classID)

    def readWidgetClassFactory(self, classID):
        if self._customWidget is not None:
            return None
        if classID is None or classID == 'None':
            return None
        classID = globals().get(classID, classID)
        if isinstance(classID, type):
            ret = classID
        elif str(classID) == 'Auto':
            ret = self.getDefaultReadWidgetClass()
        else:
            ret = TaurusWidgetFactory().getWidgetClass(classID)

        if self._compact:
            R = ret
            W = self.writeWidgetClassFactory(
                self.writeWidgetClassID, ignoreCompact=True)
            if W is None:
                return R
            switcherClass = self.getSwitcherClass()
            switcherClass.readWClass = R
            switcherClass.writeWClass = W
            return switcherClass
        return ret

    def writeWidgetClassFactory(self, classID, ignoreCompact=False):
        if self._customWidget is not None:
            return None
        if classID is None or classID == 'None':
            return None
        if self._compact and not ignoreCompact:
            return None
        classID = globals().get(classID, classID)
        if isinstance(classID, type):
            return classID
        elif str(classID) == 'Auto':
            return self.getDefaultWriteWidgetClass()
        else:
            return TaurusWidgetFactory().getWidgetClass(classID)

    def unitsWidgetClassFactory(self, classID):
        if self._customWidget is not None:
            return None
        if classID is None or classID == 'None':
            return None
        classID = globals().get(classID, classID)
        if isinstance(classID, type):
            return classID
        elif str(classID) == 'Auto':
            return self.getDefaultUnitsWidgetClass()
        else:
            return TaurusWidgetFactory().getWidgetClass(classID)

    def customWidgetClassFactory(self, classID):
        if classID is None or classID == 'None':
            return None
        classID = globals().get(classID, classID)
        if isinstance(classID, type):
            return classID
        elif str(classID) == 'Auto':
            return self.getDefaultCustomWidgetClass()
        else:
            return TaurusWidgetFactory().getWidgetClass(classID)

    def extraWidgetClassFactory(self, classID):
        if self._customWidget is not None:
            return None
        if classID is None or classID == 'None':
            return None
        classID = globals().get(classID, classID)
        if isinstance(classID, type):
            return classID
        elif str(classID) == 'Auto':
            return self.getDefaultExtraWidgetClass()
        else:
            return TaurusWidgetFactory().getWidgetClass(classID)

    def updateLabelWidget(self):
        # get the class for the widget and replace it if necessary
        klass = self.labelWidgetClassFactory(self.labelWidgetClassID)
        self._labelWidget = self._newSubwidget(self._labelWidget, klass)

        # take care of the layout
        self.addLabelWidgetToLayout()

        if self._labelWidget is not None:
            # give the new widget a reference to its buddy TaurusValue object
            self._labelWidget.taurusValueBuddy = weakref.ref(self)

            # tweak the new widget
            if self.minimumHeight() is not None:
                self._labelWidget.setMinimumHeight(self.minimumHeight())

            # set the model for the subwidget
            if hasattr(self._labelWidget, 'setModel'):
                self._labelWidget.setModel(self.getFullModelName())

    def updateReadWidget(self):
        # get the class for the widget and replace it if necessary
        try:
            klass = self.readWidgetClassFactory(self.readWidgetClassID)
            self._readWidget = self._newSubwidget(self._readWidget, klass)
        except Exception as e:
            self._destroyWidget(self._readWidget)
            self._readWidget = Qt.QLabel('[Error]')
            msg = 'Error creating read widget:\n' + str(e)
            self._readWidget.setToolTip(msg)
            self.debug(msg)
            # self.traceback(30) #warning level=30

        # take care of the layout
        self.addReadWidgetToLayout()

        if self._readWidget is not None:
            # give the new widget a reference to its buddy TaurusValue object
            self._readWidget.taurusValueBuddy = weakref.ref(self)
            if isinstance(self._readWidget, TaurusReadWriteSwitcher):
                self._readWidget.readWidget.taurusValueBuddy = weakref.ref(
                    self)
                self._readWidget.writeWidget.taurusValueBuddy = weakref.ref(
                    self)

            # tweak the new widget
            if self.minimumHeight() is not None:
                self._readWidget.setMinimumHeight(self.minimumHeight())

            # set the model for the subwidget
            if hasattr(self._readWidget, 'setModel'):
                self._readWidget.setModel(self.getFullModelName())

    def updateWriteWidget(self):
        # get the class for the widget and replace it if necessary
        klass = self.writeWidgetClassFactory(self.writeWidgetClassID)
        self._writeWidget = self._newSubwidget(self._writeWidget, klass)

        # take care of the layout
        # this is needed because the writeWidget affects to the readWritget
        # layout
        self.addReadWidgetToLayout()
        self.addWriteWidgetToLayout()

        if self._writeWidget is not None:
            # give the new widget a reference to its buddy TaurusValue object
            self._writeWidget.taurusValueBuddy = weakref.ref(self)

            # tweak the new widget
            # hide getPendingOperations of the writeWidget so that containers don't get duplicate lists
            #self._writeWidget._getPendingOperations = self._writeWidget.getPendingOperations
            #self._writeWidget.getPendingOperations = lambda : []
            self._writeWidget.valueChangedSignal.connect(self.updatePendingOpsStyle)
            self._writeWidget.setDangerMessage(self.getDangerMessage())
            self._writeWidget.setForceDangerousOperations(
                self.getForceDangerousOperations())
            if self.minimumHeight() is not None:
                self._writeWidget.setMinimumHeight(self.minimumHeight())

            # set the model for the subwidget
            if hasattr(self._writeWidget, 'setModel'):
                self._writeWidget.setModel(self.getFullModelName())

    def updateUnitsWidget(self):
        # get the class for the widget and replace it if necessary
        klass = self.unitsWidgetClassFactory(self.unitsWidgetClassID)
        self._unitsWidget = self._newSubwidget(self._unitsWidget, klass)

        # take care of the layout
        self.addUnitsWidgetToLayout()

        if self._unitsWidget is not None:
            # give the new widget a reference to its buddy TaurusValue object
            self._unitsWidget.taurusValueBuddy = weakref.ref(self)
            # tweak the new widget
            if self.minimumHeight() is not None:
                self._unitsWidget.setMinimumHeight(self.minimumHeight())

            # set the model for the subwidget
            if hasattr(self._unitsWidget, 'setModel'):
                self._unitsWidget.setModel(self.getFullModelName())

    def updateCustomWidget(self):
        # get the class for the widget and replace it if necessary
        klass = self.customWidgetClassFactory(self.customWidgetClassID)
        self._customWidget = self._newSubwidget(self._customWidget, klass)

        # take care of the layout
        self.addCustomWidgetToLayout()

        if self._customWidget is not None:
            # set the model for the subwidget
            if hasattr(self._customWidget, 'setModel'):
                self._customWidget.setModel(self.getFullModelName())

    def updateExtraWidget(self):
        # get the class for the widget and replace it if necessary
        klass = self.extraWidgetClassFactory(self.extraWidgetClassID)
        self._extraWidget = self._newSubwidget(self._extraWidget, klass)

        # take care of the layout
        self.addExtraWidgetToLayout()

        if self._extraWidget is not None:
            # give the new widget a reference to its buddy TaurusValue object
            self._extraWidget.taurusValueBuddy = weakref.ref(self)

            # set the model for the subwidget
            if hasattr(self._extraWidget, 'setModel'):
                self._extraWidget.setModel(self.getFullModelName())

    def addLabelWidgetToLayout(self):

        if self._labelWidget is not None and self.parent() is not None:
            alignment = getattr(self._labelWidget, 'layoutAlignment',
                                Qt.Qt.AlignmentFlag(0))
            self.parent().layout().addWidget(self._labelWidget, self._row, 1, 1,
                                             1, alignment)

    def addReadWidgetToLayout(self):
        if self._readWidget is not None and self.parent() is not None:
            alignment = getattr(self._readWidget, 'layoutAlignment',
                                Qt.Qt.AlignmentFlag(0))
            if self._writeWidget is None:
                self.parent().layout().addWidget(self._readWidget, self._row,
                                                 2, 1, 2, alignment)
            else:
                self.parent().layout().addWidget(self._readWidget, self._row,
                                                 2, 1, 1, alignment)

    def addWriteWidgetToLayout(self):
        if self._writeWidget is not None and self.parent() is not None:
            alignment = getattr(self._writeWidget, 'layoutAlignment',
                                Qt.Qt.AlignmentFlag(0))
            self.parent().layout().addWidget(self._writeWidget, self._row,
                                             3, 1, 1, alignment)

    def addUnitsWidgetToLayout(self):
        if self._unitsWidget is not None and self.parent() is not None:
            alignment = getattr(self._unitsWidget, 'layoutAlignment',
                                Qt.Qt.AlignmentFlag(0))
            self.parent().layout().addWidget(self._unitsWidget, self._row,
                                             4, 1, 1, alignment)

    def addCustomWidgetToLayout(self):
        if self._customWidget is not None and self.parent() is not None:
            alignment = getattr(self._customWidget, 'layoutAlignment',
                                Qt.Qt.AlignmentFlag(0))
            self.parent().layout().addWidget(self._customWidget, self._row,
                                             1, 1, -1, alignment)

    def addExtraWidgetToLayout(self):
        parent = self.parent()
        if parent is not None:
            if self._extraWidget is None:
                # Adding this spacer is some voodoo magic to avoid bug #142
                # See: http://sf.net/p/tauruslib/tickets/142/
                parent.layout().addItem(Qt.QSpacerItem(0, 0), self._row, 5)
            else:
                alignment = getattr(self._extraWidget, 'layoutAlignment',
                                    Qt.Qt.AlignmentFlag(0))
                parent.layout().addWidget(self._extraWidget, self._row,
                                          5, 1, 1, alignment)

    @Qt.pyqtSlot('QString')
    def parentModelChanged(self, parentmodel_name):
        """Invoked when the parent model changes

        :param parentmodel_name: (str) the new name of the parent model
        """
        TaurusBaseWidget.parentModelChanged(self, parentmodel_name)
        if not self._designMode:  # in design mode, no subwidgets are created
            self.updateCustomWidget()
            self.updateLabelWidget()
            self.updateReadWidget()
            self.updateWriteWidget()
            self.updateUnitsWidget()
            self.updateExtraWidget()

    @Qt.pyqtSlot('QString', name='setLabelWidget')
    def setLabelWidgetClass(self, classID):
        '''substitutes the current widget by a new one. classID can be one of:
        None, 'Auto', a TaurusWidget class name, or any class'''
        self.labelWidgetClassID = classID
        self.updateLabelWidget()

    def getLabelWidgetClass(self):
        return self.labelWidgetClassID

    def resetLabelWidgetClass(self):
        self.labelWidgetClassID = 'Auto'

    @Qt.pyqtSlot('QString', name='setReadWidget')
    def setReadWidgetClass(self, classID):
        '''substitutes the current widget by a new one. classID can be one of:
        None, 'Auto', a TaurusWidget class name, or any class'''
        self.readWidgetClassID = classID
        self.updateReadWidget()

    def getReadWidgetClass(self):
        return self.readWidgetClassID

    def resetReadWidgetClass(self):
        self.readWidgetClassID = 'Auto'

    @Qt.pyqtSlot('QString', name='setWriteWidget')
    def setWriteWidgetClass(self, classID):
        '''substitutes the current widget by a new one. classID can be one of:
        None, 'Auto', a TaurusWidget class name, or any class'''
        self.writeWidgetClassID = classID
        self.updateWriteWidget()

    def getWriteWidgetClass(self):
        return self.writeWidgetClassID

    def resetWriteWidgetClass(self):
        self.writeWidgetClassID = 'Auto'

    @Qt.pyqtSlot('QString', name='setUnitsWidget')
    def setUnitsWidgetClass(self, classID):
        '''substitutes the current widget by a new one. classID can be one of:
        None, 'Auto', a TaurusWidget class name, or any class'''
        self.unitsWidgetClassID = classID
        self.updateUnitsWidget()

    def getUnitsWidgetClass(self):
        return self.unitsWidgetClassID

    def resetUnitsWidgetClass(self):
        self.unitsWidgetClassID = 'Auto'

    @Qt.pyqtSlot('QString', name='setCustomWidget')
    def setCustomWidgetClass(self, classID):
        '''substitutes the current widget by a new one. classID can be one of:
        None, 'Auto', a TaurusWidget class name, or any class'''
        self.customWidgetClassID = classID
        self.updateCustomWidget()

    def getCustomWidgetClass(self):
        return self.customWidgetClassID

    def resetCustomWidgetClass(self):
        self.customWidgetClassID = 'Auto'

    @Qt.pyqtSlot('QString', name='setExtraWidget')
    def setExtraWidgetClass(self, classID):
        '''substitutes the current widget by a new one. classID can be one of:
        None, 'Auto', a TaurusWidget class name, or any class'''
        self.extraWidgetClassID = classID
        self.updateExtraWidget()

    def getExtraWidgetClass(self):
        return self.extraWidgetClassID

    def resetExtraWidgetClass(self):
        self.extraWidgetClassID = 'Auto'

    def setCompact(self, compact):

        # don't do anything if it is already done
        if compact == self._compact:
            return

        #do not switch to compact mode if the write widget is None
        if compact and self.writeWidget() is None:
            self.debug('No write widget. Ignoring setCompact(True)')
            return

        # Backup the current RW format
        rw = self.readWidget(followCompact=True)
        format = rw.getFormat()

        self._compact = compact
        if self.getModel():
            self.updateReadWidget()
            self.updateWriteWidget()

        # Apply the format to the new RW
        rw = self.readWidget(followCompact=True)
        rw.setFormat(format)

    def isCompact(self):
        return self._compact

    def isReadOnly(self):
        if not self.getAllowWrite():
            return True
        modelObj = self.getModelObj()
        if modelObj is None:
            return False
        return not modelObj.isWritable()

    def getModelClass(self):
        return self.__modelClass

#    def destroy(self):
#        if not self._designMode:
#            for w in [self._labelWidget, self._readWidget, self._writeWidget, self._unitsWidget, self._extraWidget]:
#                if isinstance(w,Qt.QWidget):
#                    w.setParent(self)   #reclaim the parental rights over subwidgets before destruction
#        Qt.QWidget.setParent(self,None)
#        Qt.QWidget.destroy(self)

    def createConfig(self, allowUnpickable=False):
        '''
        extending  :meth:`TaurusBaseWidget.createConfig` to store also the class names for subwidgets

        :param alllowUnpickable:  (bool)

        :return: (dict<str,object>) configurations (which can be loaded with :meth:`applyConfig`).

        .. seealso: :meth:`TaurusBaseWidget.createConfig`, :meth:`applyConfig`
        '''
        configdict = TaurusBaseWidget.createConfig(
            self, allowUnpickable=allowUnpickable)
        # store the subwidgets classIDs and configs
        for key in ('LabelWidget', 'ReadWidget', 'WriteWidget', 'UnitsWidget', 'CustomWidget', 'ExtraWidget'):
            # calls self.getLabelWidgetClass, self.getReadWidgetClass,...
            classID = getattr(self, 'get%sClass' % key)()
            if (isinstance(classID, string_types)
                    or allowUnpickable):
                #configdict[key] = classID
                configdict[key] = {'classid': classID}
                widget = getattr(self, key[0].lower() + key[1:])()
                if isinstance(widget, BaseConfigurableClass):
                    configdict[key]['delegate'] = widget.createConfig()
            else:
                self.info('createConfig: %s not saved because it is not Pickable (%s)' % (
                    key, str(classID)))

        return configdict

    def applyConfig(self, configdict, **kwargs):
        """extending :meth:`TaurusBaseWidget.applyConfig` to restore the subwidget's classes

        :param configdict: (dict)

        .. seealso:: :meth:`TaurusBaseWidget.applyConfig`, :meth:`createConfig`
        """
        # first do the basic stuff...
        TaurusBaseWidget.applyConfig(self, configdict, **kwargs)
        # restore the subwidgets classIDs
        for key in ('LabelWidget', 'ReadWidget', 'WriteWidget', 'UnitsWidget', 'CustomWidget', 'ExtraWidget'):
            if key in configdict:
                widget_configdict = configdict[key]
                getattr(self, 'set%sClass' % key)(
                    widget_configdict.get('classid', None))
                if 'delegate' in widget_configdict:
                    widget = getattr(self, key[0].lower() + key[1:])()
                    if isinstance(widget, BaseConfigurableClass):
                        widget.applyConfig(widget_configdict[
                                           'delegate'], **kwargs)

    @Qt.pyqtSlot('QString')
    def setModel(self, model):
        """extending :meth:`TaurusBaseWidget.setModel` to change the modelclass
        dynamically and to update the subwidgets"""
        self.__modelClass = taurus.Manager().findObjectClass(model or '')
        TaurusBaseWidget.setModel(self, model)
        if not self._designMode:  # in design mode, no subwidgets are created
            self.updateCustomWidget()
            self.updateLabelWidget()
            self.updateReadWidget()
            self.updateWriteWidget()
            self.updateUnitsWidget()
            self.updateExtraWidget()

    def handleEvent(self, evt_src, evt_type, evt_value):
        """Reimplemented from :meth:`TaurusBaseWidget.handleEvent`
        to update subwidgets on config events
        """
        if self._designMode:
            return

        if self.__error or evt_type == TaurusEventType.Config:
            self.updateCustomWidget()
            self.updateLabelWidget()
            self.updateReadWidget()
            self.updateWriteWidget()
            self.updateUnitsWidget()
            self.updateExtraWidget()

        # set/unset the error flag
        self.__error = (evt_type == TaurusEventType.Error)

    def isValueChangedByUser(self):
        try:
            return self._writeWidget.isValueChangedByUser()
        except AttributeError:
            return False

    def setDangerMessage(self, dangerMessage=None):
        TaurusBaseWidget.setDangerMessage(self, dangerMessage)
        try:
            return self._writeWidget.setDangerMessage(dangerMessage)
        except AttributeError:
            pass

    def setForceDangerousOperations(self, yesno):
        TaurusBaseWidget.setForceDangerousOperations(self, yesno)
        try:
            return self._writeWidget.setForceDangerousOperations(yesno)
        except AttributeError:
            pass

    def hasPendingOperations(self):
        '''self.getPendingOperations will always return an empty list, but still
        self.hasPendingOperations will look at the writeWidget's operations.
        If you want to ask the TaurusValue for its pending operations, call
        self.writeWidget().getPendingOperations()'''
        w = self.writeWidget(followCompact=True)
        if w is None:
            return False
        return w.hasPendingOperations()

    def updatePendingOpsStyle(self):
        if self._labelWidget is None:
            return
        if self.hasPendingOperations():
            newstyle = '%s {border-style: solid ; border-width: 1px; ' \
                       'border-color: blue; color: blue; border-radius:4px;}' \
                       % self._labelWidget.__class__.__name__
            oldstyle = self._labelWidget.styleSheet()
            if newstyle != oldstyle:
                self._labelWidget.setStyleSheet(newstyle)
        else:
            newstyle = '%s {border-style: solid; border-width: 1px; ' \
                       'border-color: transparent; color: black;  ' \
                       'border-radius:4px;}' % self._labelWidget.__class__.__name__
            oldstyle = self._labelWidget.styleSheet()
            if newstyle != oldstyle:
                self._labelWidget.setStyleSheet(newstyle)

    def getLabelConfig(self):
        return self._labelConfig

    @Qt.pyqtSlot('QString')
    def setLabelConfig(self, config):
        """Sets fragment configuration to the label widget.

        :param config: fragment
        :type config: str
        """
        self._labelConfig = config
        # backwards compatibility: this method used to work for setting
        # an arbitrary text to the label widget
        try:
            self.getModelFragmentObj(config)
            self._labelWidget._permanentText = None
        except Exception:
            try:
                for old in re.findall('<.+?>', config):
                    new = self._BCK_COMPAT_TAGS.get(old, old)
                    self.deprecated(dep=old, alt=new)
                    config = config.replace(old, new)

                self._labelWidget.setText(config)
            except:
                self.debug("Setting permanent text to the label widget failed")
            return

        self.updateLabelWidget()

    def resetLabelConfig(self):
        self._labelConfig = '{attr.label}'
        self.updateLabelWidget()

    def getSwitcherClass(self):
        '''Returns the TaurusValue switcher class (used in compact mode).
        Override this method if you want to use a custom switcher in
        TaurusValue subclasses.
        '''
        class TVSwitcher(TaurusReadWriteSwitcher):
            pass
        return TVSwitcher

    @classmethod
    def getQtDesignerPluginInfo(cls):
        return None
        #ret = TaurusBaseWidget.getQtDesignerPluginInfo()
        #ret['module'] = 'taurus.qt.qtgui.panel'
        #ret['icon'] = "designer:label.png"
        # return ret

    ########################################################
    # Qt properties (for designer)
    model = Qt.pyqtProperty(
        "QString", TaurusBaseWidget.getModel,  setModel, TaurusBaseWidget.resetModel)
    preferredRow = Qt.pyqtProperty(
        "int", getPreferredRow, setPreferredRow, resetPreferredRow)
    labelWidgetClass = Qt.pyqtProperty(
        "QString", getLabelWidgetClass, setLabelWidgetClass, resetLabelWidgetClass)
    readWidgetClass = Qt.pyqtProperty(
        "QString", getReadWidgetClass, setReadWidgetClass, resetReadWidgetClass)
    writeWidgetClass = Qt.pyqtProperty(
        "QString", getWriteWidgetClass, setWriteWidgetClass, resetWriteWidgetClass)
    unitsWidgetClass = Qt.pyqtProperty(
        "QString", getUnitsWidgetClass, setUnitsWidgetClass, resetUnitsWidgetClass)
    extraWidgetClass = Qt.pyqtProperty(
        "QString", getExtraWidgetClass, setExtraWidgetClass, resetExtraWidgetClass)
    labelConfig = Qt.pyqtProperty(
        "QString", getLabelConfig, setLabelConfig, resetLabelConfig)
    allowWrite = Qt.pyqtProperty(
        "bool", getAllowWrite, setAllowWrite, resetAllowWrite)
    modifiableByUser = Qt.pyqtProperty("bool", TaurusBaseWidget.isModifiableByUser,
                                       TaurusBaseWidget.setModifiableByUser, TaurusBaseWidget.resetModifiableByUser)


class TaurusValuesFrame(TaurusFrame):
    '''This is a container specialized into containing TaurusValue widgets.
    It should be used Only for TaurusValues'''

    _model = []

    @Qt.pyqtSlot('QStringList')
    def setModel(self, model):
        self._model = model
        for tv in self.getTaurusValues():
            tv.destroy()
        for m in self._model:
            taurusvalue = TaurusValue(self, self.designMode)
            taurusvalue.setMinimumHeight(20)
            taurusvalue.setModel(str(m))
            taurusvalue.setModifiableByUser(self.isModifiableByUser())

    def getModel(self):
        return self._model

    def resetModel(self):
        self.setModel([])

    def getTaurusValueByIndex(self, index):
        '''returns the TaurusValue item at the given index position'''
        return self.getTaurusValues()[index]

    def getTaurusValues(self):
        '''returns the list of TaurusValue Objects contained by this frame'''
        return [obj for obj in self.children() if isinstance(obj, TaurusValue)]

    @classmethod
    def getQtDesignerPluginInfo(cls):
        '''we don't want this widget in designer'''
        return None

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # QT properties
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    model = Qt.pyqtProperty("QStringList", getModel, setModel, resetModel)


#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
if __name__ == "__main__":

    import sys

    from taurus.qt.qtgui.application import TaurusApplication

    app = TaurusApplication(sys.argv)
    form = Qt.QMainWindow()
    # ly=Qt.QVBoxLayout(form)
    # container=Qt.QWidget()
#    container = TaurusValuesFrame()
    from taurus.qt.qtgui.panel import TaurusForm
    container = TaurusForm()
    # ly.addWidget(container)
    form.setCentralWidget(container)

    try:
        from taurus.qt.qtgui.extra_pool import PoolMotorSlim, PoolChannel
        container.setCustomWidgetMap({'SimuMotor': PoolMotorSlim,
                                      'Motor': PoolMotorSlim,
                                      'PseudoMotor': PoolMotorSlim,
                                      'PseudoCounter': PoolChannel,
                                      'CTExpChannel': PoolChannel,
                                      'ZeroDExpChannel': PoolChannel,
                                      'OneDExpChannel': PoolChannel,
                                      'TwoDExpChannel': PoolChannel})
    except:
        pass

    # set a model list
    if len(sys.argv) > 1:
        models = sys.argv[1:]
        container.setModel(models)
    else:
        models = []

    # models assigned for debugging.... comment out when releasing
    # models=['bl97/pc/dummy-01/Current','bl97/pysignalsimulator/1/value1','bl97/pc/dummy-02/RemoteMode','bl97/pc/dummy-02/CurrentSetpoint','bl97/pc/dummy-02/Current']
    # models=['bl97/pc/dummy-01/Current']
    # models=['bl97/pc/dummy-01/CurrentSetpoint','bl97/pc/dummy-02/Current','bl97/pc/dummy-02/RemoteMode','bl97/pysignalsimulator/1/value1']
    # models=['bl97/pc/dummy-01/CurrentSetpoint','bl97/pc/dummy-02/RemoteMode']
    # models=['sys/tg_test/1/state','sys/tg_test/1/status','sys/tg_test/1/short_scalar','sys/tg_test/1']
    #models =  ['sys/tg_test/1']+['sys/tg_test/1/%s_scalar'%s for s in ('float','short','string','long','boolean') ]
    # models =  ['sys/tg_test/1/float_scalar','sys/tg#_test/1/double_scalar']

    container.setModel(models)

    # container.getItemByIndex(0).writeWidget().setDangerMessage('BOOO') #uncomment to test the dangerous operation support
    # container.getItemByIndex(0).readWidget().setShowState(True)
    # container.getItemByIndex(0).setWriteWidgetClass(TaurusValueLineEdit)
    # container[0].setWriteWidgetClass('None')
    # container.setModel(models)

    container.setModifiableByUser(True)
    form.show()

    # show an Attributechooser dialog if no model was given
    if models == []:
        from taurus.qt.qtgui.panel import TaurusModelChooser
        modelChooser = TaurusModelChooser()
        modelChooser.updateModels.connect(container.setModel)
        modelChooser.show()

    sys.exit(app.exec_())



#!/usr/bin/env python

import sys
from taurus.external.qt import Qt
from taurus.qt.qtgui.plot import TaurusPlot, CurveAppearanceProperties
from taurus.qt.qtgui.application import TaurusApplication
app = TaurusApplication(sys.argv)
##########################
# BEGIN EXAMPLE CODE
##########################

import numpy
from taurus.external.qt import Qwt5

panel = TaurusPlot()

rawdata1 = {"y": 5 * numpy.random.random(10), "name": "Random"}
rawdata2 = {"x": [1, 2, 5, 7], "y": [2, 3, 1, 4], "name": "Hand-written"}
rawdata3 = {"x": numpy.arange(0, 10, 0.1), "f(x)": "sqrt(x)"}

p1 = CurveAppearanceProperties(sStyle=Qwt5.QwtSymbol.Rect,
                               sSize=5,
                               sColor="green",
                               sFill=False,
                               lStyle=Qt.Qt.NoPen)

p2 = CurveAppearanceProperties(sStyle=Qwt5.QwtSymbol.Triangle,
                               sSize=8,
                               sColor="red",
                               sFill=True,
                               lColor="red",
                               lStyle=Qt.Qt.DashLine)

panel.attachRawData(rawdata1, properties=p1)
panel.attachRawData(rawdata2, properties=p2)
panel.attachRawData(rawdata3)


########################
# END EXAMPLE CODE
########################
panel.show()
sys.exit(app.exec_())



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""This module provides basic utilities for usage in any test"""

__docformat__ = 'restructuredtext'

import functools


def insertTest(klass=None, helper_name=None, test_method_name=None,
               test_method_doc=None, tested_name=None, test_skip=None,
               **helper_kwargs):
    """Decorator that inserts test methods from a helper method that accepts
    arguments.
    `insertTest` provides a very economic API for creating new tests for a given
    class based on a helper method.
    `insertTest` accepts the following arguments:

      - helper_name (str): the name of the helper method. `insertTest` will
                           insert a test method which calls the helper with
                           any the helper_kwargs (see below).
      - test_method_name (str): Optional. Name of the test method to be used.
                                If None given, one will be generated from the
                                tested class and helper names.
      - test_method_doc (str): Optional. The docstring for the inserted test
                               method (this shows in the unit test output).
                               If None given, a default one is generated which
                               includes the input parameters and the helper
                               name.
      - tested_name (str): Optional. The name of the class or feature being
                           tested (if given, it will be used in default method
                           names and docstrings).

      - test_skip (str): Optional. A reason for skipping the test. If None
                         given, the test will not be skipped

      - \*\*helper_kwargs: All remaining keyword arguments are passed to the
                           helper.

    This decorator can be considered a "base" decorator. It is often used to
    create other decorators in which the helper method is pre-set, as in
    the following example::

        isPos = functools.partial(insertTest, helper_name='isPositive')

        @isPos(x=2)
        @isPos(x=10)
        class Foo(unittest.TestCase):
            def isPositive(self, x):
                self.assertTrue(x > 0)

    """
    # Recipe to support decorating with and without arguments
    if klass is None:
        return functools.partial(insertTest, helper_name=helper_name,
                                 test_method_name=test_method_name,
                                 test_method_doc=test_method_doc,
                                 tested_name=tested_name,
                                 test_skip=test_skip,
                                 **helper_kwargs)

    # Check arguments and provide defaults
    if helper_name is None:
        raise ValueError('helper_name argument is not optional')

    if test_method_name is None:
        test_method_name = 'test_'
        if tested_name:
            test_method_name += '%s_' % tested_name
        test_method_name += helper_name
    # Append an index if necessary to avoid overwriting existing test methods
    name, i = test_method_name, 1
    while (hasattr(klass, name)):
        i += 1
        name = "%s_%i" % (test_method_name, i)
    test_method_name = name

    if test_method_doc is None:
        argsrep = ', '.join(['%s=%s' % (k, repr(v))
                             for k, v in helper_kwargs.items()])
        if tested_name:
            test_method_doc = 'Testing %s with %s(%s)' % (tested_name,
                                                          helper_name, argsrep)
        else:
            test_method_doc = 'Testing %s(%s)' % (helper_name, argsrep)

    # New test implementation
    def newTest(obj):
        helper = getattr(obj, helper_name)
        return helper(**helper_kwargs)

    # Add the custom docstring
    newTest.__doc__ = test_method_doc

    # Skip the test if the test_skip kwarg was passed
    if test_skip is not None:
        import unittest
        newTest = unittest.skip(test_skip)(newTest)

    # Add the new test method with the new implementation
    setattr(klass, test_method_name, newTest)

    return klass


if __name__ == '__main__':

    # a demo of use of insertTest

    import unittest

    isPos = functools.partial(insertTest, helper_name='isPositive')
    isNeg = functools.partial(insertTest, helper_name='isPositive',
                              expected=False)

    @isPos
    @isPos(x=2)
    @isPos(x=10)
    @isPos(x=5)
    @isNeg(x=-1)
    class FooTest(unittest.TestCase):

        def isPositive(self, x=1, expected=True):
            self.assertEqual(x > 0, expected)

    unittest.main(verbosity=2)



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""
emitter.py: This module provides a task scheduler used by TaurusGrid and 
    TaurusDevTree widgets
"""

from future import standard_library
standard_library.install_aliases()
from builtins import str
from builtins import object
from queue import Queue, Empty
import traceback

from future.utils import string_types

import taurus
from taurus.external.qt import Qt
from taurus.core.util.log import Logger


###############################################################################
# Helping methods


def isString(seq):
    if isinstance(seq, string_types):
        return True  # It matches most python str-like classes
    if any(s in str(type(seq)).lower() for s in ('vector', 'array', 'list',)):
        return False
    if 'qstring' == str(type(seq)).lower():
        return True  # It matches QString
    return False


def modelSetter(obj, model):
    """
    This class is used for convenience as TaurusEmitterThread standard method
    """
    # print 'In modelSetter(%s,%s)' % (str(obj),str(model))
    if hasattr(obj, 'setModel') and model is not None:
        obj.setModel(model)
    return


class MethodModel(object):
    """
    Class to emulate method execution as a setModel
    """

    def __init__(self, method):
        self.method = method

    def setModel(self, value):
        return self.method(value)


class QEmitter(Qt.QObject):
    """Emitter class providing two signals."""

    doSomething = Qt.pyqtSignal(list)
    somethingDone = Qt.pyqtSignal()
    newQueue = Qt.pyqtSignal()


###############################################################################


class TaurusEmitterThread(Qt.QThread):
    """
    The TaurusEmitterThread Class
    ==========================

    This object get items from a python Queue and performs a thread safe 
    operation on them.
    It is useful to serialize Qt tasks in a background thread.

    :param parent: a Qt/Taurus object
    :param name: identifies object logs
    :param queue: if None parent.getQueue() is used, if not then the queue 
        passed as argument is used
    :param method: the method to be executed using each queue item as argument
    :param cursor: if True or QCursor a custom cursor is set while 
        the Queue is not empty

    How TaurusEmitterThread works
    --------------------------

    TaurusEmitterThread is a worker thread that processes a queue of iterables 
    passed as arguments to the specified method every time that  
    ``doSomething()`` is called:

     * ``self.method(*item)`` will be called if TaurusEmitterThread.method 
        has been initialized.
     * ``item[0](item[1:])`` will be called if ``method`` is not initialized 
        and the first element of the queued iterable is *callable*.

    TaurusEmitterThread uses two queues:

     * ``self.queue`` manages the objects added externally:

      + the ``next()`` method passes objects from ``self.queue`` 
        to ``self.todo queue``
      + every time that a *somethingDone* signal arrives ``next()`` is called.
      + ``next()`` can be called also externally to ensure that the main queue 
      is being processed.
      + the queue can be accessed externally using ``getQueue()``
      + ``getQueue().qsize()`` returns number of remaining objects in queue.
      + while there are objects in queue the ``.next()`` method will 
        override applications cursor. a call to ``next()`` with an empty queue 
        will restore the original cursor.

     * ``self.todo`` is managed by the ``run()/start()`` method:

      - a loop waits continuously for new objects in ``self.todo`` queue.
      - if an object is found, it is sent in a *doSomething* signal.
      - if *"exit"* is found the loop exits.

    Usage example
    -------------

    .. code-block:: python

        #Applying TaurusEmitterThread to an existing class:
        from queue import Queue
        from functools import partial

        def modelSetter(args):
            obj,model = args[0],args[1]
            obj.setModel(model)

        klass TaurusGrid(Qt.QFrame, TaurusBaseWidget):
            ...
            def __init__(self, parent = None, designMode = False):
                ...
                self.modelsQueue = Queue()
                self.modelsThread = TaurusEmitterThread(parent=self,
                        queue=self.modelsQueue,method=modelSetter )
                ...
            def build_widgets(...):
                ...
                            previous,synoptic_value = \
                                synoptic_value,TaurusValue(cell_frame)
                            #synoptic_value.setModel(model)
                            self.modelsQueue.put((synoptic_value,model))
                ...
            def setModel(self,model):
                ...
                if hasattr(self,'modelsThread') and \
                        not self.modelsThread.isRunning():
                    self.modelsThread.start()
                elif self.modelsQueue.qsize():
                    self.modelsThread.next()
                ...

    """

    def __init__(self, parent=None, name='', queue=None, method=None,
                 cursor=None, sleep=5000, polling=0, loopwait=5):
        """
        Parent must be not None and must be a TaurusGraphicsScene!

        :param queue: pass an external action queue (optional)
        :param method: action processor (e.g. modelSetter)
        :param cursor: QCursor during process (optional)
        :param sleep: delay in ms before thread start
        :param polling: process actions at fix period (milliseconds)
        :param loopwait: wait N milliseconds between actions
        """
        Qt.QThread.__init__(self, parent)
        self.name = name
        self.log = Logger('TaurusEmitterThread(%s)' % self.name)
        self.log.setLogLevel(self.log.Info)
        self.queue = queue or Queue()
        self.todo = Queue()
        self.method = method
        self.cursor = Qt.QCursor(
            Qt.Qt.WaitCursor) if cursor is True else cursor
        self._cursor = False
        self.timewait = int(sleep)
        self.polling = int(polling)
        self.loopwait = int(loopwait)
        if self.polling:
            self.refreshTimer = Qt.QTimer()
            self.refreshTimer.timeout.connect(self.onRefresh)
        else:
            self.refreshTimer = None

        self.emitter = QEmitter()
        self.emitter.moveToThread(Qt.QApplication.instance().thread())
        # Mandatory!!! if parent is set before changing thread it could lead to
        # segFaults!
        self.emitter.setParent(Qt.QApplication.instance())
        self._done = 0
        # Moved to the end to prevent segfaults ...
        self.emitter.doSomething.connect(self._doSomething)

        if not self.refreshTimer:
            self.emitter.somethingDone.connect(self.next)

    def onRefresh(self):
        try:
            size = self.getQueue().qsize()
            if size:
                self.log.info('onRefresh(%s)' % size)
                self.next()
            else:
                self.log.debug('onRefresh()')
        except:
            self.log.warning(traceback.format_exc())

    def getQueue(self):
        if self.queue:
            return self.queue
        elif hasattr(self.parent(), 'getQueue'):
            self.parent().getQueue()
        else:
            return None

    def getDone(self):
        """ Returns % of done tasks in 0-1 range """
        pending = self.getQueue().qsize()
        return float(self._done) / (self._done + pending)

    def clear(self):
        while not self.todo.empty():
            self.todo.get()
        while not self.getQueue().empty():
            self.getQueue().get()
            self._done += 1

    def purge(self, obj):
        """
        Remove a given object from all queues
        """
        nqueue = Queue()
        while not self.todo.empty():
            i = self.todo.get()
            if obj not in i:
                nqueue.put(i)
        while not self.queue.empty():
            i = self.queue.get()
            if obj not in i:
                nqueue.put(i)
        while not nqueue.empty():
            self.queue.put(nqueue.get())
        self.next()

    def _doSomething(self, params):
        self.log.debug('At TaurusEmitterThread._doSomething(%s)' % str(params))
        if not self.method:
            method, args = params[0], params[1:]
        else:
            method, args = self.method, params
        if method:
            try:
                method(*args)
            except:
                self.log.error('At TaurusEmitterThread._doSomething(%s): \n%s'
                               % (list(map(str, args)), traceback.format_exc()))
        self.emitter.somethingDone.emit()
        self._done += 1
        return

    def next(self):
        queue = self.getQueue()
        msg = ('At TaurusEmitterThread.next(), %d items remaining.'
               % queue.qsize())
        if (queue.empty() and not self.polling):
            self.log.info(msg)
        else:
            self.log.debug(msg)
        try:
            if not queue.empty():
                if not self._cursor and self.cursor is not None:
                    Qt.QApplication.instance().setOverrideCursor(
                        Qt.QCursor(self.cursor))
                    self._cursor = True
                # A blocking get here would hang the GUIs!!!
                item = queue.get(False)
                self.todo.put(item)
                self.log.debug('Item added to todo queue: %s' % str(item))
            elif self._cursor:
                Qt.QApplication.instance().restoreOverrideCursor()
                self._cursor = False

        except Empty:
            self.log.warning(traceback.format_exc())
            pass
        except:
            self.log.warning(traceback.format_exc())
        return

    def run(self):
        Qt.QApplication.instance().thread().msleep(self.timewait)
        self.log.info('#' * 80)
        self.log.info('At TaurusEmitterThread.run()')
        self.next()

        if self.refreshTimer:
            self.refreshTimer.start(self.polling)

        while True:
            self.log.debug('At TaurusEmitterThread.run() loop.')
            item = self.todo.get(True)
            if isString(item):
                if item == "exit":
                    break
                else:
                    continue
            self.log.debug('Emitting doSomething signal ...')
            self.emitter.doSomething.emit(item)
            if self.loopwait:
                self.msleep(self.loopwait)
                # End of while
        self.log.info(
            '#' * 80 + '\nOut of TaurusEmitterThread.run()' + '\n' + '#' * 80)
        # End of Thread


class DelayedSubscriber(Logger):
    """
    DelayedSubscriber(schema) will use a TaurusEmitterThread to perform
    a thread safe delayed subscribing on all Attributes of a given 
    Taurus Schema that has not been previously subscribed.

    .. warning:: This class belongs to a "Delayed Event Subscription" API added
                 in v.4.2.1-alpha as an *experimental* feature. This API may
                 not be stable and/or it may be removed in a future release
                 (even on a minor version change)
    """

    def __init__(self, schema, parent=None, sleep=10000, pause=5, period=0):

        self._schema = schema
        self.call__init__(Logger, 'DelayedSubscriber(%s)' % self._schema, None)
        self._factory = taurus.Factory(schema)

        self._modelsQueue = Queue()
        self._modelsThread = TaurusEmitterThread(parent=parent,
                                                 queue=self._modelsQueue,
                                                 method=self._modelSubscriber,
                                                 sleep=sleep, loopwait=pause,
                                                 polling=period)

        self._modelsQueue.put((self.addUnsubscribedAttributes,))
        self._modelsThread.start()

    def _modelSubscriber(self, method, args=[]):
        self.debug('modelSubscriber(%s,%s)' % (method, args))
        return method(*args)

    def getUnsubscribedAttributes(self):
        """Check all pending subscriptions in the current factory
        """
        attrs = []
        for name, attr in self._factory.getExistingAttributes().items():
            if attr is None:
                continue
            elif attr.hasListeners() and not attr.isUsingEvents():
                attrs.append(attr)

        return attrs

    def addUnsubscribedAttributes(self):
        """Schedule subscription for all pending attributes
        """
        try:
            items = self.getUnsubscribedAttributes()
            if len(items):
                self.info('addUnsubscribedAttributes([%d])' % len(items))
                for attr in items:
                    self._addModelObj(attr)
                self._modelsThread.next()
                self.info('Thread queue: [%d]' % (self._modelsQueue.qsize()))
        except:
            self.warning(traceback.format_exc())

    def _addModelObj(self, modelObj):
        parent = modelObj.getParentObj()
        if parent:
            proxy = parent.getDeviceProxy()
            if not proxy:
                self.debug('addModelObj(%s), proxy not available' % modelObj)
                return

        self._modelsQueue.put((modelObj.subscribePendingEvents,))
        self.debug('addModelObj(%s)' % str(modelObj))

    def cleanUp(self):
        self.trace("[DelayedSubscriber] cleanUp")
        self._modelsThread.stop()
        Logger.cleanUp(self)


class SingletonWorker(object):
    """
    SingletonWorker is used to manage TaurusEmitterThread as Singleton objects

    SingletonWorker is constructed using the same arguments 
    than TaurusTreadEmitter ; but instead of creating a QThread for each 
    instance it creates a single QThread for all instances.

    The Queue is still different for each of the instances; it is connected 
    to the TaurusEmitterThread signals (*next()* and *somethingDone()*) 
    and each Worker queue is used as a feed for the shared QThread.

    This implementation reduced the cpu of vacca application in a 50% factor.

    :param parent: a Qt/Taurus object
    :param name: identifies object logs
    :param queue: if None parent.getQueue() is used, if not then the queue 
        passed as argument is used
    :param method: the method to be executed using each queue item as argument
    :param cursor: if True or QCursor a custom cursor is set while 
        the Queue is not empty
    """
    _thread = None

    def __init__(self, parent=None, name='', queue=None, method=None,
                 cursor=None, sleep=5000, log=Logger.Warning, start=True):
        self.name = name
        self.log = Logger('SingletonWorker(%s)' % self.name)
        self.log.setLogLevel(log)
        self.log.info('At SingletonWorker.__init__(%s)' % self.name)
        self.parent = parent
        self.method = method
        self._running = False
        if SingletonWorker._thread is None:
            SingletonWorker._thread = TaurusEmitterThread(
                parent, name='SingletonWorker', cursor=cursor, sleep=sleep)
        self.thread = SingletonWorker._thread
        self.queue = queue or Queue()
        if start:
            self.start()

    def put(self, item, block=True, timeout=None):
        self.getQueue().put(item, block, timeout)

    def size(self):
        return self.getQueue().qsize()

    def next(self, item=None):
        if item is not None:
            self.put(item)
        elif self.queue.empty():
            return
        msg = ('At SingletonWorker.next(), '
               '%d items not passed yet to Emitter.'
               % self.queue.qsize())
        self.log.info(msg)
        # (queue.empty() and self.log.info or self.log.debug)(msg)
        try:
            i = 0
            while not self.queue.empty():
                # A blocking get here would hang the GUIs!!!
                item = self.queue.get(False)
                if self.method:
                    self.thread.getQueue().put([self.method] + list(item))
                else:
                    self.thread.getQueue().put(item)
                i += 1
            self.log.info('%d Items added to emitter queue' % i)
            self.thread.emitter.newQueue.emit()
        except Empty:
            self.log.warning(traceback.format_exc())
        except:
            self.log.warning(traceback.format_exc())
        return

    def getQueue(self):
        return self.queue

    def getDone(self):
        return self.thread.getDone()

    def start(self):
        self.thread.emitter.somethingDone.connect(self.next)
        self.thread.emitter.newQueue.connect(self.thread.next)
        try:
            self.thread.start()
        except:
            pass
        self.next()
        self._running = True
        return

    def stop(self):
        self.thread.emitter.somethingDone.disconnect(self.next)
        self.thread.emitter.newQueue.disconnect(self.thread.next)
        self._running = False
        return

    def clear(self):
        """
        This method will clear queue only if next() has not been called.
        If you call self.thread.clear() it will clear objects for all workers!,
        be careful
        """
        while not self.queue.empty():
            self.queue.get()
            # self.thread.clear()

    def purge(self, obj):
        """
        Remove a given object from all queues
        """
        nqueue = Queue()
        while not self.queue.empty():
            i = self.queue.get()
            if obj not in i:
                nqueue.put(i)
        while not nqueue.empty():
            self.queue.put(nqueue.get())
        self.next()

    def isRunning(self):
        return self._running

    def isFinished(self):
        return self.thread.isFinished()

    def finished(self):
        return self.thread.finished()

    def started(self):
        return self._running

    def terminated(self):
        return self.thread.terminated()

    def sleep(self, s):
        return self.thread.sleep(s)



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""Test for taurus.core.taurushelper"""

#__all__ = []

__docformat__ = 'restructuredtext'

import numpy
import functools
import unittest
from taurus.core.units import Quantity
import taurus
from taurus.test import insertTest
from taurus.core import (TaurusElementType, TaurusAttribute, TaurusDevice,
                         TaurusAuthority, DataType, TaurusAttrValue, DataFormat)

# decorators
isValid = functools.partial(insertTest, helper_name='isValid')
isInvalid = functools.partial(insertTest, helper_name='isValid',
                              expected=False)
isValidAuth = functools.partial(insertTest, helper_name='isValid',
                                elementType=TaurusElementType.Authority)
isInvalidAuth = functools.partial(insertTest, helper_name='isValid',
                                  expected=False,
                                  elementType=TaurusElementType.Authority)
isValidDev = functools.partial(insertTest, helper_name='isValid',
                               elementType=TaurusElementType.Device)
isInvalidDev = functools.partial(insertTest, helper_name='isValid',
                                 expected=False,
                                 elementType=TaurusElementType.Device)
isValidAttr = functools.partial(insertTest, helper_name='isValid',
                                elementType=TaurusElementType.Attribute)
isInvalidAttr = functools.partial(insertTest, helper_name='isValid',
                                  expected=False,
                                  elementType=TaurusElementType.Attribute)

from taurus import tauruscustomsettings
DEFAULT_SCHEME = getattr(tauruscustomsettings, 'DEFAULT_SCHEME')


@isValidAuth(name='//implicittangodb:123', skip=(DEFAULT_SCHEME != 'tango'))
@isValidAuth(name='tango://foo:10000')
@isInvalidAuth(name='tango://foo bar:10000')
@isInvalidAuth(name='tango://db')
@isInvalidAuth(name='tango://db:10000/a/b/c')
@isInvalidAuth(name='tango://db:')
@isInvalidAuth(name='tango://db:10000/')
@isInvalidAuth(name='tango://a/b/c')
@isValidDev(name='implicit/tango/devname', skip=(DEFAULT_SCHEME != 'tango'))
@isValidDev(name='implicittangoalias', skip=(DEFAULT_SCHEME != 'tango'))
@isValidDev(name='tango://db:10000/alias')
@isValidDev(name='tango://foo:10000/a/b/c')
@isValidDev(name='tango:alias')
@isValidDev(name='tango://alias', strict=False)
@isValidDev(name='tango:a/b/c')
@isValidDev(name='tango:a/b  /c')
@isValidDev(name='tango://a/b/c', strict=False)
@isInvalidDev(name='tango://a/b/c', strict=True)
@isInvalidDev(name='tango://alias', strict=True)
@isInvalidDev(name='tango:a/b/c/')
@isInvalidDev(name='tango:a/b')
@isInvalidDev(name='tango:a/b/c/d')
@isValidAttr(name='implicittangoalias/attrname',
             skip=(DEFAULT_SCHEME != 'tango'))
@isValidAttr(name='implicit/tango/attr/name', skip=(DEFAULT_SCHEME != 'tango'))
@isValidAttr(name='tango:alias/attr')
@isValidAttr(name='tango://alias/attr', strict=False)
@isValidAttr(name='tango://foo:10000/a/b/c/d')
@isValidAttr(name='tango:a/b/c/d')
@isValidAttr(name='tango://a/b/c/d', strict=False)
# @isValidAttr(name='epics://my:example.RBV') disabled till bug #215 is fixed
@isValidAttr(name='tango:a/b  /c/d')
@isInvalidAttr(name='tango:a/b/c/d/')
@isInvalidAttr(name='tango:a/b/c')
@isValidAttr(name='implicit/tango/attr/name#label',
             skip=(DEFAULT_SCHEME != 'tango'))
@isValidAttr(name='implicittangoalias/attrname#label',
             skip=(DEFAULT_SCHEME != 'tango'))
@isValidAttr(name='tango:alias/attr#label')
@isValidAttr(name='tango://foo:10000/a/b/c/d#label')
@isValidAttr(name='tango:a/b/c/d#label')
@isValidAttr(name='tango:a/b  /c/d#label')
@isValidAttr(name='tango://a/b/c/d?configuration=label', strict=False)
@isInvalidAttr(name='tango://a/b/c/d#label', strict=True)
@isInvalidAttr(name='tango://a/b')
@isInvalidAttr(name='tango://a/b/c')
@isInvalidAttr(name='tango://a/b/c/d')
@isValidAuth(name='eval://localhost')
@isInvalidAuth(name='eval://foo')
@isInvalidAuth(name='evaluation://db=foo')
@isInvalidAuth(name='evaluation://localhost')
@isInvalidAuth(name='eval://db=foo')
@isInvalidAuth(name='eval://localhost#')
@isInvalidAuth(name='eval://')
@isInvalidAuth(name='evaluation://user@host:port')
@isInvalidAuth(name='eval://user@host:port')
@isInvalidAuth(name='eval://user:host@port')
@isInvalidAuth(name='eval://user@host:port/1')
@isInvalidAuth(name='eval:{tango://localhost:10000}')
@isValidAuth(name='evaluation://db=foo', strict=False)
@isValidAuth(name='eval://db=foo', strict=False)
@isInvalidAuth(name='eval://foo', strict=False)
@isInvalidAuth(name='eval://DB=foo', strict=False)
@isValidDev(name='eval:@module.EvalClass')
@isValidDev(name='eval:@1')
@isValidDev(name='eval:@Foo')
@isValidDev(name='eval:@FOO')
@isValidDev(name='eval://localhost/@Foo')
@isValidDev(name='eval://localhost/@mod.Class')
@isInvalidDev(name='evaluation:@Foo')
@isInvalidDev(name='eval://db=foo')
@isInvalidDev(name='eval://dev=bar')
@isInvalidDev(name='eval:')
@isInvalidDev(name='eval:@Foo#')
@isInvalidDev(name='eval:1')
@isInvalidDev(name='eval:DEV=bar')
@isInvalidDev(name='eval:@DEV=bar')
@isValidDev(name='eval://db=foo;dev=bar', strict=False)
@isValidDev(name='eval://dev=bar', strict=False)
@isValidDev(name='eval://dev=bar?a=1;b=2', strict=False)
# yes, this is ok
@isValidDev(name='eval://dev=bar?configuration', strict=False)
@isInvalidDev(name='eval:dev=bar', strict=False)
@isInvalidDev(name='eval://db=Bla/dev=bar', strict=False)
@isValidAttr(name='eval:1')
@isValidAttr(name='eval:/1')
@isValidAttr(name='eval:@bar/1')
@isValidAttr(name='evaluation://db=foo;dev=bar;1', strict=False)
@isValidAttr(name='eval://db=foo;dev=bar;1', strict=False)
@isValidAttr(name='eval://dev=bar;foo', strict=False)
@isValidAttr(name='eval://2+3', strict=False)
@isValidAttr(name='evaluation://2+3', strict=False)
@isValidAttr(name='eval://a+b?a=2;b=3', strict=False)
@isValidAttr(name='eval://a*{tango://a/b/c/d}?a=10', strict=False)
@isValidAttr(name='eval://dev=bar/1', strict=False)
@isValidAttr(name='eval://a[:2]?a=[1,2,3]', strict=False)
@isValidAttr(name='eval://{eval://1}', strict=False)
@isValidAttr(name='eval://{eval://{eval://1}}', strict=False)
@isValidAttr(name='eval:{tango:a/b/c/d}*2')
@isValidAttr(name='eval:k=2;a={tango:a/b/c/d};a*k')
@isValidAttr(name='eval:@mymod.MyEvalClass/k=2;a={tango:a/b/c/d};a*k')
@isValidAttr(name='eval://localhost/k=2;a={tango:a/b/c/d};a*k')
@isValidAttr(name='eval://localhost/@Foo/k=2;a={tango:a/b/c/d};a*k')
@isValidAttr(name='eval:a=[1,2,3];a[:2]')
@isValidAttr(name='eval:{eval:1}')
@isValidAttr(name='eval:{eval:{eval:1}}')
@isInvalidAttr(name='EVAL:1')
@isInvalidAttr(name='evaluation:1')
@isInvalidAttr(name='eval://db=foo;dev=bar;1')
@isInvalidAttr(name='eval://2+3')
@isInvalidAttr(name='evaluation://2+3')
@isInvalidAttr(name='eval:a+b?a=2;b=3')
@isValidAttr(name='eval:foo#')
@isInvalidAttr(name='eval:{tango:a/b/c}*2')
@isInvalidAttr(name='eval:k-a;k=2;a=3')
@isInvalidAttr(name='eval:a=1;b=2;a+b;a*2')
@isInvalidAttr(name='eval:a=1;b=2;a+b;;;')
@isInvalidAttr(name='eval:a=1;;;b=2;a+b')
@isValidAttr(name='eval:a=2;b=3;a+b#label')
@isValidAttr(name='eval:{tango:a/b/c/d}#')
@isValidAttr(name='eval:{tango:a/b/c/d}*2#label')
@isValidAttr(name='eval:k=2;a=3;a*k#units')
@isValidAttr(name='eval://localhost/@mymod.MyClass/1#label')
@isValidAttr(name='eval:{tango:a/b/c/d}*2#1+2')
@isInvalidAttr(name='eval:{tango:a/b/c}*2#')
@isInvalidAttr(name='eval:2# ')
@isInvalidAttr(name='evaluation:1#label')
@isInvalidAttr(name='eval://dev=bar;foo#label')
@isInvalidAttr(name='eval:a+b;a=2;b=3#label')
@isInvalidAttr(name='eval://foo')
@isValidAttr(name='evaluation://db=foo;dev=bar;1?configuration=label', strict=False)
@isValidAttr(name='eval://db=foo;dev=bar;1?configuration=label', strict=False)
@isValidAttr(name='eval://dev=bar;foo?configuration=label', strict=False)
@isValidAttr(name='eval://2+3?configuration=label', strict=False)
@isValidAttr(name='eval://a+b?a=2;b=3?configuration=label', strict=False)
class IsValidNameTestCase(unittest.TestCase):
    '''Class to test validity or invalidity of taurus URIs'''

    def isValid(self, name=None, expected=True, elementType=None,
                strict=True, skip=False, skipmsg='Test explicitly disabled'):
        '''Helper method to test validity or invalidity of taurus URIs'''
        if skip:
            self.skipTest(skipmsg)
        if elementType is None:
            elementName = '<any>'
        else:
            elementName = TaurusElementType.whatis(elementType)
            elementType = [elementType]
        manager = taurus.Manager()
        scheme = manager.getScheme(name)
        supportedSchemes = list(manager.getPlugins().keys())
        if scheme not in supportedSchemes:
            self.skipTest('"%s" scheme not supported' % scheme)
        returned = taurus.isValidName(name, etypes=elementType, strict=strict)

        msg = ('isValidName returns %s (expected %s for name=%s and etype=%s)' %
               (returned, expected, name, elementName))
        self.assertEqual(returned, expected, msg)

    def test_unsupported(self):
        '''Testing that an unsupported scheme validates as False'''
        returned = taurus.isValidName('_unsupported_://foo')
        msg = ('Validating unsupported schemes must return False (got %s)' %
               returned)
        self.assertFalse(returned, msg)


#@insertTest(helper_name='get_object', name='eval://foo')
@insertTest(helper_name='get_object', name='eval://localhost')
class AuthorityTestCase(unittest.TestCase):
    '''TestCase for the taurus.Authority helper'''

    def get_object(self, name=None, klass=None):
        '''check if Authority returns the expected type'''
        if klass is None:
            klass = TaurusAuthority
        manager = taurus.Manager()
        scheme = manager.getScheme(name)
        supportedSchemes = list(manager.getPlugins().keys())
        if scheme not in supportedSchemes:
            self.skipTest('"%s" scheme not supported' % scheme)
        a = taurus.Authority(name)
        msg = ('%s("%s") is not an instance of %s (it is %s)' %
               (taurus.Authority.__name__, name,
                klass.__name__, a.__class__.__name__))
        self.assertTrue(isinstance(a, klass), msg)


@insertTest(helper_name='get_object', name='tango:a/b/c')
@insertTest(helper_name='get_object', name='eval://localhost/@Foo')
@insertTest(helper_name='get_object', name='eval:@Foo')
@insertTest(helper_name='get_object', name='eval://dev=Foo')
@insertTest(helper_name='get_object', name='eval:@datetime.*')
@insertTest(helper_name='get_object', name='eval:@d=datetime.date(2017,3,29)')
class DeviceTestCase(unittest.TestCase):
    '''TestCase for the taurus.Device helper'''

    def get_object(self, name=None, klass=None):
        '''check if Device returns the expected type'''
        if klass is None:
            klass = TaurusDevice
        manager = taurus.Manager()
        scheme = manager.getScheme(name)
        supportedSchemes = list(manager.getPlugins().keys())
        if scheme not in supportedSchemes:
            self.skipTest('"%s" scheme not supported' % scheme)

        a = taurus.Device(name)
        msg = ('%s("%s") is not an instance of %s (it is %s)' %
               (taurus.Device.__name__, name,
                klass.__name__, a.__class__.__name__))

        self.assertTrue(isinstance(a, klass), msg)


@insertTest(helper_name='read_attr',
            name='eval:2*{eval:x=2;x}',
            expected=dict(label='2*x', data_format=DataFormat._0D,
                          type=DataType.Integer))
@insertTest(helper_name='read_attr',
            name='eval:x=-1;x+{eval:x=2;x}+{eval:x=10;x}',
            expected=dict(data_format=DataFormat._0D, wvalue=None, w_value=None,
                          label='x+x+x',
                          type=DataType.Integer))
@insertTest(helper_name='read_attr', name='eval://2*{eval://1.0}+{eval://2.0}',
            expected=dict(data_format=DataFormat._0D, type=DataType.Float,
                          label='2*1.0+2.0'))
@insertTest(helper_name='read_attr',
            name='eval:2*{tango:sys/tg_test/1/short_scalar}',
            expected=dict(data_format=DataFormat._0D, type=DataType.Integer,
                          label='2*short_scalar'))
@insertTest(helper_name='read_attr',
            name='eval:1==1',
            expected=dict(data_format=DataFormat._0D, type=DataType.Boolean,
                          label='1==1'))
@insertTest(helper_name='read_attr',
            name='eval:[1,2,3]',
            expected=dict(data_format=DataFormat._1D, type=DataType.Integer,
                          label='[1,2,3]'))
@insertTest(helper_name='read_attr',
            name='eval:img={tango:sys/tg_test/1/short_image_ro};img+10*rand(*img.shape)',
            expected=dict(data_format=DataFormat._2D, type=DataType.Float))
@insertTest(helper_name='read_attr',
            name='eval:{tango:sys/tg_test/1/string_image}',
            expected=dict(data_format=DataFormat._2D, type=DataType.String))
@insertTest(helper_name='get_object', name='tango:a/b/c/d')
@insertTest(helper_name='read_attr', name='tango:sys/tg_test/1/short_scalar',
            expected=dict(data_format=DataFormat._0D, type=DataType.Integer,
                          label='short_scalar'))
@insertTest(helper_name='read_attr', name='tango:sys/tg_test/1/float_image',
            expected=dict(data_format=DataFormat._2D, type=DataType.Float,
                          label='float_image'))
@insertTest(helper_name='read_attr', name='tango:sys/tg_test/1/boolean_scalar',
            expected=dict(data_format=DataFormat._0D, type=DataType.Boolean,
                          label='boolean_scalar'))
@insertTest(helper_name='read_attr', name='tango:sys/tg_test/1/double_spectrum',
            expected=dict(data_format=DataFormat._1D, type=DataType.Float,
                          label='double_spectrum'))
@insertTest(helper_name='read_attr', name='eval:1.0',
            expected=dict(rvalue=1.0, value=1.0,
                          wvalue=None, w_value=None, label='1.0',
                          type=DataType.Float, data_format=DataFormat._0D))
@insertTest(helper_name='read_attr', name='eval:x=3;x+1',
            expected=dict(rvalue=4, value=4, wvalue=None, w_value=None,
                          label='x+1',
                          type=DataType.Integer, data_format=DataFormat._0D))
@insertTest(helper_name='read_attr', name='eval:2*{eval:3}',
            expected=dict(rvalue=6, value=6, label='2*3',
                          wvalue=None, w_value=None,
                          type=DataType.Integer, data_format=DataFormat._0D))
@insertTest(helper_name='read_attr', name='eval:2*{eval:3*{eval:4}}',
            expected=dict(rvalue=24, value=24, label='2*3*4',
                          wvalue=None, w_value=None, type=DataType.Integer,
                          data_format=DataFormat._0D))
@insertTest(helper_name='read_attr', name='eval:1*{eval:2*{eval:3*{eval:4}}}',
            expected=dict(rvalue=24, value=24,
                          label='1*2*3*4',
                          wvalue=None, w_value=None, type=DataType.Integer,
                          data_format=DataFormat._0D))
@insertTest(helper_name='read_attr', name='eval:{eval:1}+{eval:9}',
            expected=dict(rvalue=10, value=10, label='1+9',
                          wvalue=None, w_value=None, type=DataType.Integer,
                          data_format=DataFormat._0D))
@insertTest(helper_name='read_attr', name='eval:"aaa"+"bbb"',
            expected=dict(rvalue="aaabbb", value="aaabbb", label='"aaa"+"bbb"',
                          wvalue=None, w_value=None, type=DataType.String,
                          data_format=DataFormat._0D))
@insertTest(helper_name='read_attr', name='eval:{eval:"aaa"}+{eval:"bbb"}',
            expected=dict(rvalue="aaabbb", value="aaabbb",
                          wvalue=None, w_value=None,
                          label='"aaa"+"bbb"',
                          type=DataType.String, data_format=DataFormat._0D))
@insertTest(helper_name='read_attr', name='eval:{eval:"a"}*3',
            expected=dict(rvalue="aaa", value="aaa", label='"a"*3',
                          wvalue=None, w_value=None, type=DataType.String,
                          data_format=DataFormat._0D))
@insertTest(helper_name='read_attr', name='eval:len("abc")',
            expected=dict(rvalue=3, wvalue=None, value=3, w_value=None,
                          label='len("abc")',
                          type=DataType.Integer, data_format=DataFormat._0D))
@insertTest(helper_name='read_attr', name='eval:"a.b.c".split(".")',
            expected=dict(rvalue=['a', 'b', 'c'], value=['a', 'b', 'c'],
                          wvalue=None, w_value=None,
                          label='"a.b.c".split(".")', type=DataType.String,
                          data_format=DataFormat._1D))
@insertTest(helper_name='read_attr', name='eval:[1,1,1].count(1)',
            expected=dict(rvalue=3, wvalue=None, value=3, w_value=None,
                          type=DataType.Integer, data_format=DataFormat._0D))
@insertTest(helper_name='read_attr', name='eval:[1,2,3].reverse()',
            expected=dict(rvalue=None, value=None, wvalue=None, w_value=None,
                          label='[1,2,3].reverse()'))
@insertTest(helper_name='read_attr',
            name='eval:[1,1,3].count(1)*{eval:[1,1,3].count(1)*{eval:2}}',
            expected=dict(rvalue=8, value=8, wvalue=None, w_value=None,
                          type=DataType.Integer))
@insertTest(helper_name='read_attr', name='eval:"split.split".split("split")',
            expected=dict(rvalue=['', '.', ''], value=['', '.', ''],
                          wvalue=None, w_value=None,
                          label='"split.split".split("split")',
                          type=DataType.String))
@insertTest(helper_name='read_attr', name='eval:Quantity(1.0)',
            expected=dict(rvalue=Quantity(1.0), value=1.0,
                          w_value=None, wvalue=None, label='Quantity(1.0)',
                          type=DataType.Float, data_format=DataFormat._0D))
@insertTest(helper_name='read_attr', name='eval:Quantity("1km")',
            expected=dict(rvalue=Quantity(1000, 'm'), value=1,
                          w_value=None, wvalue=None, label='Quantity("1km")',
                          type=DataType.Integer, data_format=DataFormat._0D))
@insertTest(helper_name='read_attr', name='eval:Quantity("0.1km")',
            expected=dict(rvalue=Quantity(100, 'm'), value=0.1,
                          w_value=None, wvalue=None, label='Quantity("0.1km")',
                          type=DataType.Float, data_format=DataFormat._0D))
@insertTest(helper_name='read_attr', name='eval:Q("1m")*3+{eval:Q(3,"dm")}',
            expected=dict(rvalue=Quantity("3.3m"), value=3.3,
                          wvalue=None, w_value=None,
                          label='Q("1m")*3+Q(3,"dm")',
                          type=DataType.Float))
@insertTest(helper_name='read_attr', name='eval:Q("1km").to("mm").magnitude',
            expected=dict(rvalue=1e6, value=1e6, wvalue=None, w_value=None,
                          label='Q("1km").to("mm").magnitude',
                          type=DataType.Float))
@insertTest(helper_name='read_attr',
            name='eval:@d=datetime.date(1931,4,14)/d.isoformat()',
            expected=dict(rvalue='1931-04-14', value='1931-04-14',
                          wvalue=None, w_value=None,
                          label='d.isoformat()',
                          type=DataType.String))
@insertTest(helper_name='read_attr', name='eval:1#')
class AttributeTestCase(unittest.TestCase):
    '''TestCase for the taurus.Attribute helper'''

    def get_object(self, name=None, klass=None):
        '''check if Attribute returns the expected type'''
        if klass is None:
            klass = TaurusAttribute
        manager = taurus.Manager()
        scheme = manager.getScheme(name)
        supportedSchemes = list(manager.getPlugins().keys())
        if scheme not in supportedSchemes:
            self.skipTest('"%s" scheme not supported' % scheme)
        a = taurus.Attribute(name)
        msg = ('%s("%s") is not an instance of %s (it is %s)' %
               (taurus.Attribute.__name__, name,
                klass.__name__, a.__class__.__name__))
        self.assertTrue(isinstance(a, klass), msg)
        return a

    def read_attr(self, name=None, expected=None, skip=False,
                  skipmsg='Test explicitly disabled'):
        '''check creation and correct reading of an attribute'''
        if skip:
            self.skipTest(skipmsg)
        if expected is None:
            expected = {}
        a = self.get_object(name)
        readvalue = a.read()
        msg = ('read() for "%s" did not return a TaurusAttrValue (got a %s)' %
               (name, readvalue.__class__.__name__))
        self.assertTrue(isinstance(readvalue, TaurusAttrValue), msg)
        for k, exp in expected.items():
            try:
                got = getattr(readvalue, k)
            except AttributeError:
                msg = ('The read value for "%s" does not provide info on %s' %
                       (name, k))
                self.fail(msg)
            msg = ('Expected %s=%r for "%s" (got %r)' % (k, exp, name, got))

            try:
                self.assertEqual(got, exp, msg)
            except ValueError:
                # Validate the arrays
                try:
                    # for those values that can be handled by numpy.allclose()
                    chk = numpy.allclose(got, exp)
                except:
                    # for the rest
                    if isinstance(got, numpy.ndarray):
                        got = got.tolist()
                    chk = bool(got == exp)

                self.assertTrue(chk, msg)


if __name__ == '__main__':
    pass



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""This module provides a set of basic taurus widgets based on QLineEdit"""

__all__ = ["TaurusConfigurationPanel", "TangoConfigLineEdit",
           "TaurusConfigLineEdit"]

__docformat__ = 'restructuredtext'

import sys

from taurus.external.qt import Qt
from taurus.qt.qtgui.util import getWidgetsOfType
from taurus.qt.qtgui.util.ui import UILoadable
from taurus.qt.qtgui.base import TaurusBaseWritableWidget
from taurus.core.taurusattribute import TaurusAttribute
from taurus.core.taurusoperation import WriteAttrOperation
from taurus.core.util.log import taurus4_deprecation


class TangoConfigLineEdit(Qt.QLineEdit, TaurusBaseWritableWidget):

    def __init__(self, qt_parent=None, designMode=False):
        name = self.__class__.__name__
        self.call__init__wo_kw(Qt.QLineEdit, qt_parent)
        self.call__init__(TaurusBaseWritableWidget,
                          name, designMode=designMode)

        self.textChanged.connect(self.notifyValueChanged)
        self.returnPressed.connect(self.writeValue)
        self.editingFinished.connect(self._onEditingFinished)

    def _onEditingFinished(self):
        if self._autoApply:
            self.writeValue()

    def handleEvent(self, evt_src, evt_type, evt_value):
        self.notifyValueChanged()

    def getModelClass(self):
        return TaurusAttribute

    def postAttach(self):
        if self.isAttached():
            try:
                v = self._getAttrInfoExMember(self.modelFragmentName)
            except:
                v = None
            self.setValue(v)

    def updatePendingOperations(self):
        model = self.getModelObj()
        try:
            model_value = self.text()
            wigdet_value = self.getValue()
            if model.areStrValuesEqual(model_value, wigdet_value):
                self._operations = []
            else:
                operation = WriteAttrOperation(model, wigdet_value,
                                               self.getOperationCallbacks())
                self._operations = [operation]
        except:
            self._operations = []
        self.updateStyle()

    def setValue(self, v):
        model = self.getModelObj()
        if model is None:
            v_str = str(v)
        else:
            v_str = str(self.getValue())
        self.blockSignals(True)
        self.setText(v_str.strip())
        self.blockSignals(False)

    def _getAttrInfoExMember(self, member):
        model_obj = self.getModelObj()
        attrInfoEx = model_obj.getAttributeInfoEx()
        if '_alarm' in member or '_warning' in member:
            alarms = getattr(attrInfoEx, 'alarms')
            return getattr(alarms, member)
        else:
            return getattr(attrInfoEx, member)

    def _setAttrInfoExMember(self, member, value):
        model_obj = self.getModelObj()
        attrInfoEx = model_obj.getAttributeInfoEx()
        if '_alarm' in member or '_warning' in member:
            alarms = getattr(attrInfoEx, 'alarms')
            setattr(alarms, member, value)
        else:
            setattr(attrInfoEx, member, value)
        model_obj.setConfigEx(attrInfoEx)

    def getValue(self):
        if self.modelFragmentName is None:
            return None
        return self._getAttrInfoExMember(self.modelFragmentName)

    def notifyValueChanged(self):
        if str(self.text()) != str(self.getValue()):
            style = 'TangoConfigLineEdit {color: %s; font-weight: %s}' %\
                    ('blue', 'bold')
            self.setStyleSheet(style)
        else:
            style = 'TangoConfigLineEdit {color: %s; font-weight: %s}' %\
                    ('black', 'normal')
            self.setStyleSheet(style)

    def writeValue(self):
        if self.modelFragmentName is not None:
            self._setAttrInfoExMember(self.modelFragmentName, str(self.text()))
        style = 'TangoConfigLineEdit {color: %s; font-weight: %s}' %\
                ('black', 'normal')
        self.setStyleSheet(style)

    def setModel(self, model):
        TaurusBaseWritableWidget.setModel(self, model)
        self.notifyValueChanged()

    @classmethod
    def getQtDesignerPluginInfo(cls):
        ret = TaurusBaseWritableWidget.getQtDesignerPluginInfo()
        ret['module'] = 'taurus.qt.qtgui.input'
        ret['icon'] = "designer:lineedit.png"
        return ret

#-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # QT properties
#-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    model = Qt.pyqtProperty("QString", TaurusBaseWritableWidget.getModel,
                            TaurusBaseWritableWidget.setModel,
                            TaurusBaseWritableWidget.resetModel)

    autoApply = Qt.pyqtProperty("bool", TaurusBaseWritableWidget.getAutoApply,
                                TaurusBaseWritableWidget.setAutoApply,
                                TaurusBaseWritableWidget.resetAutoApply)

    forcedApply = Qt.pyqtProperty("bool", TaurusBaseWritableWidget.getForcedApply,
                                  TaurusBaseWritableWidget.setForcedApply,
                                  TaurusBaseWritableWidget.resetForcedApply)


class TaurusConfigLineEdit(TangoConfigLineEdit):

    @taurus4_deprecation(alt='TangoConfigLineEdit')
    def __init__(self, qt_parent=None, designMode=False):
        TangoConfigLineEdit.__init__(self, qt_parent=None, designMode=False)

    @classmethod
    def getQtDesignerPluginInfo(cls):
        return None


@UILoadable(with_ui='_ui')
class TaurusConfigurationPanel(Qt.QWidget):

    def __init__(self, parent=None, designMode=False):
        Qt.QWidget.__init__(self, parent)
        self.loadUi()

        self._ui.pushButtonOk.clicked.connect(self._onOk)
        self._ui.pushButtonApply.clicked.connect(self._onApply)
        self._ui.pushButtonCancel.clicked.connect(self._onCancel)
        self._ui.pushButtonRestore.clicked.connect(self._onRestore)

    def _onOk(self):
        self._onApply()
        self._onCancel()

    def _onApply(self):
        widgets = getWidgetsOfType(self, TangoConfigLineEdit)
        for w in widgets:
            param = str(w.text())
            if w.getValue() != param:
                w.writeValue()

    def _onCancel(self):
        self.close()

    def _onRestore(self):
        widgets = getWidgetsOfType(self, TangoConfigLineEdit)
        for w in widgets:
            param = str(w.text())
            if w.getValue() != param:
                w.setText(param)

    def setModel(self, model):
        self._ui.fullNameLineEdit.setText(model)
        model += '#'
        self._ui.labelConfig.setModel(model + 'label')
        self._ui.unitConfig.setModel(model + 'unit')
        self._ui.displayUnitConfig.setModel(model + 'display_unit')
        self._ui.standardUnitConfig.setModel(model + 'standard_unit')
        self._ui.formatConfig.setModel(model + 'format')
        self._ui.descriptionConfig.setModel(model + 'description')
        self._ui.valueMinConfig.setModel(model + 'min_value')
        self._ui.valueMaxConfig.setModel(model + 'max_value')
        self._ui.alarmMinConfig.setModel(model + 'min_alarm')
        self._ui.alarmMaxConfig.setModel(model + 'max_alarm')
        self._ui.warningMinConfig.setModel(model + 'min_warning')
        self._ui.warningMaxConfig.setModel(model + 'max_warning')


def main():
    import sys
    attr_name = sys.argv[1]
    a = Qt.QApplication([])
    w1 = TaurusConfigurationPanel()
    w1._ui.pushButtonOk.setVisible(True)
    w1._ui.pushButtonCancel.setVisible(True)
    w1.setModel(attr_name)
    w1.show()
    return a.exec_()

if __name__ == "__main__":
    sys.exit(main())
